/* allocate.c -- GC glue for builtin types
** 
** Copyright (c) 2009 Ian Piumarta
** All Rights Reserved
** 
** Permission is hereby granted, free of charge, to any person obtaining a
** copy of this software and associated documentation files (the 'Software'),
** to deal in the Software without restriction, including without limitation
** the rights to use, copy, modify, merge, publish, distribute, and/or sell
** copies of the Software, and to permit persons to whom the Software is
** furnished to do so, provided that the above copyright notice(s) and this
** permission notice appear in all copies of the Software.  Inclusion of the
** the above copyright notice(s) and this permission notice in supporting
** documentation would be appreciated but is not required.
** 
** THE SOFTWARE IS PROVIDED 'AS IS'.  USE ENTIRELY AT YOUR OWN RISK.
** 
** Last edited: 2010-08-03 11:14:43 by piumarta on emilia-2.local
*/

static oop *default_threadSpecificLocation(void);

typedef oop *(*fptsl)(void);

fptsl threadSpecificLocation= default_threadSpecificLocation;

#define threadSpecific (*(*threadSpecificLocation)())

#if USE_NIL
# if USE_TAG
    static inline void MARK(oop o)	{ if (!((long)(o)&1)) GC_mark(o); }
# else
    static inline void MARK(oop o)	{ GC_mark(o); }
# endif
#else
# if USE_TAG
    static inline void MARK(oop o)	{ if ((o) && !((long)(o)&1)) GC_mark(o); }
# else
    static inline void MARK(oop o)	{ if (o) GC_mark(o); }
# endif
#endif

static void markFunction(void *ptr)
{
  oop obj= (oop)ptr;
  MARK(getProperties(obj));
#if USE_OT
  void *body= obj->_body;
  if (!body) return;
  GC_mark_leaf(body);
#endif
  get(getType(obj), Type,mark)(obj);
}

static void mark_Undefined(oop obj)	{}
static void mark_Tag(oop obj)		{}
static void mark_Long(oop obj)		{}
static void mark_Float(oop obj)		{}
static void mark_Pointer(oop obj)	{ MARK(get(obj, Pointer,name));  MARK(get(obj, Pointer,type)); }
static void mark_Pair(oop obj)		{ MARK(get(obj, Pair,head));  MARK(get(obj, Pair,tail)); }

static void mark_String(oop obj)
{
  char *elements= get(obj, String,elements);
  if (elements) GC_mark_leaf(elements);
}

static void mark_Symbol(oop obj)	{ MARK(get(obj, Symbol,string)); }

static void mark_Array(oop obj)
{
  MARK(get(obj, Array,undefined));
  oop  sizo=     get(obj, Array,size);
  oop *elements= get(obj, Array,elements);
  if (elements) GC_mark_leaf(elements);
  if (notNil(sizo)) {
    long size= getLong(sizo), index;
    assert(elements);
    for (index= 0;  index < size;  ++index) MARK(elements[index]);
  }
}

static void mark_ByteArray(oop obj)
{
  byte *elements= get(obj, ByteArray,elements);
  if (elements) GC_mark_leaf(elements);
}

static void mark_WordArray(oop obj)
{
  word *elements= get(obj, WordArray,elements);
  if (elements) GC_mark_leaf(elements);
}

static void mark_Syntax(oop obj)
{
  MARK(get(obj, Syntax,form));
  MARK(get(obj, Syntax,compiler));
}

static void mark_Expr(oop obj)
{
  MARK(get(obj, Expr,expr));
  MARK(get(obj, Expr,params));
  MARK(get(obj, Expr,body));
  MARK(get(obj, Expr,env));
  MARK(get(obj, Expr,method));
  MARK(get(obj, Expr,name));
}

static void mark_Subr(oop obj)		{ MARK(get(obj, Subr,name)); }

static void mark_Type(oop obj) {
  MARK(get(obj, Type,name));
  MARK(get(obj, Type,fields));
}

static void mark_Context(oop obj) {
  MARK(get(obj, Context,pc));
  MARK(get(obj, Context,env));
  MARK(get(obj, Context,caller));
  MARK(get(obj, Context,callee));
  MARK(get(obj, Context,function));
  MARK(get(obj, Context,kernel));
  void *sb= get(obj, Context,stackBase);
  if (sb) GC_mark_leaf(sb);
  oop  *sp= get(obj, Context,sp);
  oop  *sl= get(obj, Context,stackLimit);
  while (sp < sl) {
    oop obj= *sp++;
    MARK(obj);
  }
}

static void mark_type(oop obj)
{
#if USE_OT
  oop *field= (oop *)obj->_body;
#else
  oop *field= (oop *)&obj->_body;
#endif
  oop  fields;
  for (fields= get(getType(obj), Type,fields);  isPair(fields);  ++field, fields= getTail(fields))
    if (s_oopAt == cadar(fields))
      MARK(*field);
}

static void mark_threadSpecificLocation(void)
{
  MARK(threadSpecific);
}

#undef MARK

#if defined(NDEBUG)
# define mkStamp(OBJECT) (OBJECT)
#else
# define mkStamp(OBJECT) _mkStamp(OBJECT, __FILE__, __LINE__, __FUNCTION__)
  static inline oop _mkStamp(oop obj, const char *file, int line, const char *func)
  {
    obj->_file= file;
    obj->_line= line;
    obj->_func= func;
    return obj;
  }
#endif

#define mkLong(L)					_mkLong(L)
#define mkPointer(ADDRESS, NAME, TYPE)		mkStamp(_mkPointer(ADDRESS, NAME, TYPE))
#define mkFloat(F)				mkStamp(_mkFloat(F))
#define mkPair(HEAD, TAIL)			mkStamp(_mkPair(HEAD, TAIL))
#define mkStringN(SIZE)				mkStamp(_mkStringN(SIZE))
#define mkString(S)				mkStamp(_mkString(S))
#define mkSymbol(S)				mkStamp(_mkSymbol(S))
#define mkArray(SIZE)				mkStamp(_mkArray(SIZE))
#define mkWordArray(SIZE)			mkStamp(_mkWordArray(SIZE))
#define mkByteArray(SIZE)			mkStamp(_mkByteArray(SIZE))
#define mkSyntax(F)				mkStamp(_mkSyntax(F))
#define mkExpr(EXPR, PARAMS, BODY, METHOD, ENV)	mkStamp(_mkExpr(EXPR, PARAMS, BODY, METHOD, ENV))
#define mkSubr(EXEC, NAME)			mkStamp(_mkSubr(EXEC, NAME))
#define mkType(NAME, FIELDS, SIZE, MARK, PRINT)	mkStamp(_mkType(NAME, FIELDS, SIZE, MARK, PRINT))
#define mkContextN(SIZE)			mkStamp(_mkContextN(SIZE))
#define mkContext(FUNC, ENV, CALLER)		mkStamp(_mkContext(FUNC, ENV, CALLER))

#if USE_OT
static oop empty(int type)
{
  oop obj= (oop)GC_malloc(sizeof(struct Object));
  obj->_type= type;
#if USE_OT
  obj->_body= 0;
#endif
  return obj;
}
#endif

static oop allocPointers(int type, size_t size)
{
#if USE_OT
  oop obj= empty(type);
  GC_PROTECT(obj);
  obj->_body= (body_t *)GC_malloc(size);
  GC_UNPROTECT(obj);
  return obj;
#else
  oop obj= (oop)GC_malloc(sizeof(struct Object) - sizeof(body_t) + size);
  obj->_type= type;
  return obj;
#endif
}

#if USE_OT
static oop allocBytes(int type, size_t size)
{
#if USE_OT
  oop obj= empty(type);
  GC_PROTECT(obj);
  obj->_body= (body_t *)GC_malloc_atomic(size);
  GC_UNPROTECT(obj);
  return obj;
#else
  oop obj= (oop)GC_malloc_atomic(sizeof(type_t) + sizeof(int) + size);
  obj->_type= type;
  return obj;
#endif
}
#endif

#define newPointers(Type)	allocPointers(i##Type, sizeof(Type))
#if USE_OT
# define newBytes(Type)		allocBytes(i##Type, sizeof(Type))
#else
# define newBytes		newPointers
#endif

static oop _mkLong(long l)
{
#if USE_TAG
  if ((l ^ (l << 1)) >= 0)
    return (oop)(l << 1 | 1);
#endif
  oop obj= newBytes(Long);
  set(obj, Long,value, l);
  return obj;
}

static oop _mkPointer(void *address, oop name, oop type)
{
  oop obj= newPointers(Pointer);
  set(obj, Pointer,address,  address);
  set(obj, Pointer,name,     name);
  set(obj, Pointer,type,     type);
  return obj;
}

static oop _mkFloat(double f)
{
  oop obj= newBytes(Float);
  set(obj, Float,value, f);
  return obj;
}

static oop _mkPair(oop head, oop tail)
{
  oop obj= newPointers(Pair);
  set(obj, Pair,head, head);
  set(obj, Pair,tail, tail);
  return obj;
}

static oop _mkStringN(long size)
{
  long capacity= size ? size : 16;
  oop obj= newBytes(String);
  GC_PROTECT(obj);
  set(obj, String,elements, 0);
  char *elements= GC_malloc_atomic(sizeof(byte) * capacity);
  set(obj, String,elements, elements);
  set(obj, String,size,     nil);
  set(obj, String,size,     mkLong(size));
  GC_UNPROTECT(obj);
  return obj;
}

static oop _mkString(char *s)
{
  long size= strlen(s) + 1;
  oop obj= mkStringN(size);
  memcpy(get(obj, String,elements), s, size);
  set(obj, String,size, mkLong(getLong(get(obj, String,size)) - 1));
  return obj;
}

static oop _mkSymbol(char *s)
{
  oop syms;
  long size= strlen(s);
  for (syms= symbols;  isPair(syms);  syms= getTail(syms)) {
    oop sym= getHead(syms);
    oop str= get(sym, Symbol,string);
    if (size == getLong(get(str, String,size)))
      if (!memcmp(s, get(str, String,elements), size))
	return sym;
  }
  oop sym= newPointers(Symbol);		GC_PROTECT(sym);
  set(sym, Symbol,string, mkString(s));
  set(sym, Symbol,opcode, 0);
  symbols= mkPair(sym, symbols);	GC_UNPROTECT(sym);
  return sym;
}

static oop _mkArray(long size)
{
  long capacity= size ? size : 16;
  oop obj= newPointers(Array);
  GC_PROTECT(obj);
  set(obj, Array,size,     nil);
  set(obj, Array,elements, 0);	/* ok to mark elements if gc in next line */
  set(obj, Array,elements, GC_malloc(sizeof(oop) * capacity));
  if (size > 0) {
    oop *elements= get(obj, Array,elements);
    long i;
    for (i= 0;  i < size;  ++i)
      elements[i]= nil;
  }
  set(obj, Array,size, mkLong(size));
  GC_UNPROTECT(obj);
  return obj;
}

static oop _mkWordArray(long size)
{
  long capacity= size ? size : 16;
  oop obj= newPointers(WordArray);
  GC_PROTECT(obj);
  set(obj, WordArray,elements, (word *)0);	/* ok to mark elements if gc in next line */
  set(obj, WordArray,elements, GC_malloc_atomic(sizeof(word) * capacity));
  set(obj, WordArray,size, nil);
  set(obj, WordArray,size, mkLong(size));
  GC_UNPROTECT(obj);
  return obj;
}

static oop _mkByteArray(long size)
{
  long capacity= size ? size : 16;
  oop obj= newPointers(ByteArray);
  GC_PROTECT(obj);
  set(obj, ByteArray,elements, (byte *)0);	/* ok to mark elements if gc in next line */
  set(obj, ByteArray,elements, GC_malloc_atomic(capacity));
  set(obj, ByteArray,size, nil);
  set(obj, ByteArray,size, mkLong(size));
  GC_UNPROTECT(obj);
  return obj;
}

static oop _mkSyntax(oop f)
{
  oop obj= newPointers(Syntax);
  set(obj, Syntax,form, f);
  set(obj, Syntax,compiler, f);
  return obj;
}

static oop _mkExpr(oop expr, oop params, oop body, oop method, oop env)
{
  oop obj= newPointers(Expr);
  set(obj, Expr,expr,   expr);
  set(obj, Expr,params, params);
  set(obj, Expr,body,   body);
  set(obj, Expr,method, method);
  set(obj, Expr,env,    env);
  set(obj, Expr,name,   nil);
  return obj;
}

static oop _mkSubr(exec_t exec, oop name)
{
  oop obj= newPointers(Subr);
  set(obj, Subr,exec, exec);
  set(obj, Subr,name, name);
  return obj;
}

static oop _mkType(oop name, oop fields, long size, mark_t mark, print_t print)
{
  long index= notNil(types) ? getLong(get(types, Array,size)) : -1;
  oop  obj= newPointers(Type);
  set(obj, Type,index,  index);
  set(obj, Type,name,   name);
  set(obj, Type,fields, fields);
  set(obj, Type,size,   size);
  set(obj, Type,mark,   mark);
  set(obj, Type,print,  print);
  if (index >= 0) {
    GC_PROTECT(obj);
    array_append(types, obj);
    GC_UNPROTECT(obj);
    assert(obj == get(types, Array,elements)[index]);
  }
  return obj;
}

#define STACK_MIN_SIZE	32

static oop _mkContextN(size_t size)
{
  oop obj= newPointers(Context);			GC_PROTECT(obj);
  set(obj, Context,pc,         nil);
  set(obj, Context,env,	       nil);
  set(obj, Context,caller,     nil);
  set(obj, Context,callee,     nil);
  set(obj, Context,function,   nil);
  set(obj, Context,kernel,     nil);
  set(obj, Context,ip,	       0);
  set(obj, Context,sp,         0);
  set(obj, Context,stackBase,  0);
  set(obj, Context,stackLimit, 0);
  oop *base=  GC_malloc(sizeof(oop) * size);
  oop *limit= base + size;
  set(obj, Context,stackBase,  base);
  set(obj, Context,stackLimit, limit);
  set(obj, Context,sp,         limit);
  set(obj, Context,kernel,     threadSpecific);
							GC_UNPROTECT(obj);
  return obj;
}

static oop _mkContext(oop function, oop env, oop caller)
{
  oop obj;
  if (notNil(caller) && notNil(obj= get(caller, Context,callee))) {
    set(obj, Context,sp,     	get(obj,    Context,stackLimit));
    set(obj, Context,kernel, 	get(caller, Context,kernel));
    set(obj, Context,function,	function);
    set(obj, Context,env,	env);
    set(obj, Context,caller,    caller);
  }
  else {
    obj= newPointers(Context);			GC_PROTECT(obj);
    set(obj, Context,pc,         nil);
    set(obj, Context,callee,     nil);
    set(obj, Context,sp,         (oop *)nil);
    set(obj, Context,stackBase,  (oop *)nil);
    set(obj, Context,stackLimit, (oop *)nil);
    oop *base=  GC_malloc(sizeof(oop) * STACK_MIN_SIZE);
    oop *limit= base + STACK_MIN_SIZE;
    set(obj, Context,stackBase,  base);
    set(obj, Context,stackLimit, limit);
    set(obj, Context,sp,         limit);
    set(obj, Context,function,	 function);
    set(obj, Context,env,	 env);
    set(obj, Context,caller,     caller);
    set(obj, Context,kernel,	 nil);
    if (notNil(caller)) {
      set(caller, Context,callee, obj);
      set(obj,    Context,kernel, get(caller, Context,kernel));
    }
    else
      set(obj, Context,kernel, threadSpecific);
						GC_UNPROTECT(obj);
  }
  return obj;
}

static inline oop mkBoolean(int arg)
{
  return arg ? t : nil;
}

static oop *default_threadSpecificLocation(void)
{
  static oop specific= nil;
  return &specific;
}
