/* debug.c -- mainly for printing builtin types
** 
** Copyright (c) 2009 Ian Piumarta
** All Rights Reserved
** 
** Permission is hereby granted, free of charge, to any person obtaining a
** copy of this software and associated documentation files (the 'Software'),
** to deal in the Software without restriction, including without limitation
** the rights to use, copy, modify, merge, publish, distribute, and/or sell
** copies of the Software, and to permit persons to whom the Software is
** furnished to do so, provided that the above copyright notice(s) and this
** permission notice appear in all copies of the Software.  Inclusion of the
** the above copyright notice(s) and this permission notice in supporting
** documentation would be appreciated but is not required.
** 
** THE SOFTWARE IS PROVIDED 'AS IS'.  USE ENTIRELY AT YOUR OWN RISK.
** 
** Last edited: 2010-08-03 11:14:42 by piumarta on emilia-2.local
*/

static int lastCycle= 0;

static oop _print(oop obj, FILE *stream, int cycle)
{
  if (!obj)
#  if USE_NIL
    fprintf(stream, "<NULL>");
#  else
    fprintf(stream, "()");
#  endif
  else {
    get(getType(obj), Type,print)(obj, stream, cycle);
  }
  return obj;
}

static oop fprint(oop obj, FILE *stream)
{
  return _print(obj, stream, /*++*/lastCycle);
}

static oop fprintsp(oop obj, FILE *stream)
{
  fprint(obj, stream);
  fprintf(stream, " ");
  return obj;
  (void)fprintsp;
}

static oop fprintln(oop obj, FILE *stream)
{
  fprint(obj, stream);
  fprintf(stream, "\n");
  return obj;
}

static oop print(oop obj)	{ return fprint  (obj, stdout); }
static oop printsp(oop obj)	{ return fprintsp(obj, stdout); }
static oop println(oop obj)	{ return fprintln(obj, stdout); }

static oop print_Undefined(oop obj, FILE *stream, int cycle)
{
  fprintf(stream, "()");
  return obj;
}

static oop print_Tag(oop obj, FILE *stream, int cycle)
{
  fprintf(stream, "@%ld", getLong(obj));
  return obj;
}

static oop print_Long(oop obj, FILE *stream, int cycle)
{
#if !defined(NDEBUG)
  if ((long)obj & 1) fprintf(stream, "+");
#endif
  fprintf(stream, "%ld", getLong(obj));
  return obj;
}

static oop print_Float(oop obj, FILE *stream, int cycle)
{
  fprintf(stream, "%lf", getFloat(obj));
  return obj;
}

static oop print_Pointer(oop obj, FILE *stream, int cycle)
{
  fprintf(stream, "#<%p", get(obj, Pointer,address));
  if (notNil(get(obj, Pointer,name))) fprintf(stream, "=%s", tostring(get(obj, Pointer,name)));
  if (notNil(get(obj, Pointer,type))) _print(get(obj, Pointer,type), stream, cycle);
  fprintf(stream, ">");
  return obj;
}

static oop print_Pair(oop obj, FILE *stream, int cycle)
{
#if 0
# warning fixme
#else
  if (isPair(obj) && s_globals == getHead(obj)) {
    fprintf(stream, "*globals*");
    obj->_cycle= cycle;
    return obj;
  }
#endif
  //#ifndef NDEBUG
  if (obj->_cycle) {
    fprintf(stream, "...");
    return obj;
  }
  obj->_cycle++;
  //#endif
  fprintf(stream, "(");
  _print(getHead(obj), stream, cycle);
  oop list;
  for (list= obj;  (list= getTail(list)) && isPair(list);) {
#  if 0
#   warning fixme
#  else
    if (isPair(getHead(list)) && s_globals == getHead(getHead(list))) {
      fprintf(stream, " *globals*)");
      goto done;
      return obj;
    }
#  endif
    //#ifndef NDEBUG
    if (list->_cycle) {
      fprintf(stream, " ...)");
      break;
    }
    //#endif
    fprintf(stream, " ");
    _print(getHead(list), stream, cycle);
  }
  if (notNil(list)) {
    fprintf(stream, " . ");
    _print(list, stream, cycle);
  }
  fprintf(stream, ")");
 done:
  obj->_cycle--;
  return obj;
}

static oop print_String(oop obj, FILE *stream, int cycle)
{
  char *s= get(obj, String,elements);
  long i= getLong(get(obj, String,size));
  //fprintf(stream, "<%s>", s);
  fprintf(stream, "\"");
  while (*s && i--) {
    int c= *s++;
    switch (c) {
    case '"':	fprintf(stream, "\\\"");	break;
    case '\\':  fprintf(stream, "\\\\");	break;
    default:
      if (' ' <= c && c <= '~') fputc(c, stream);
      else fprintf(stream, "\\%03o", c);
    }
  }
  fprintf(stream, "\"");
  return obj;
}

static oop print_Symbol(oop obj, FILE *stream, int cycle)
{
  //fprintf(stream, "#\"%s\"", get(get(obj, Symbol,string), String,elements));
  fprintf(stream, "%s", get(get(obj, Symbol,string), String,elements));
  return obj;
}

static oop print_Array(oop obj, FILE *stream, int cycle)
{
  long sz= getLong(get(obj, Array,size)), i;
  fprintf(stream, "Array(%ld/%ld:", sz, (long)(GC_size(get(obj, Array,elements)) / sizeof(oop)));
  for (i= 0;  i < sz;  ++i) {
    fprintf(stream, " ");
    _print(get(obj, Array,elements)[i], stream, cycle);
  }
  fprintf(stream, ")");
  return obj;
}

static oop print_ByteArray(oop obj, FILE *stream, int cycle)
{
  long sz= getLong(get(obj, ByteArray,size)), i;
  fprintf(stream, "ByteArray(%ld/%ld:", sz, (long)(GC_size(get(obj, ByteArray,elements))));
  for (i= 0;  i < sz;  ++i)
    fprintf(stream, " %02x", get(obj, ByteArray,elements)[i]);
  fprintf(stream, ")");
  return obj;
}

static oop print_WordArray(oop obj, FILE *stream, int cycle)
{
  long sz= getLong(get(obj, WordArray,size)), i;
  fprintf(stream, "WordArray(%ld/%ld:", sz, (long)(GC_size(get(obj, WordArray,elements)) / sizeof(word)));
  for (i= 0;  i < sz;  ++i)
    fprintf(stream, " %08lx", get(obj, WordArray,elements)[i]);
  fprintf(stream, ")");
  return obj;
}

static oop print_Syntax(oop obj, FILE *stream, int cycle)
{
  fprintf(stream, "#Syntax(");
  _print(get(obj, Syntax,form), stream, cycle);
  fprintf(stream, " ");
  _print(get(obj, Syntax,compiler), stream, cycle);
  fprintf(stream, ")");
  return obj;
}

static oop print_Expr(oop obj, FILE *stream, int cycle)
{
  fprintf(stream, "#Expr(");
#if 1
  if (notNil(get(obj, Expr,name)))
    _print(get(obj, Expr,name), stream, cycle);
  else
    _print(get(obj, Expr,params), stream, cycle);
  //fprintf(stream, " ");
  //_print(get(obj, Expr,body), stream, cycle);
#else
  if (isNil(get(obj, Expr,method))) {
    fprintf(stream, " ");
    _print(get(obj, Expr,body), stream, cycle);
  } else {
    _print(get(obj, Expr,expr), stream, cycle);
    fprintf(stream, "\n");
    dump("", get(obj, Expr,method));
  }
#endif
  fprintf(stream, ")");
  return obj;
}

static oop print_Subr(oop obj, FILE *stream, int cycle)
{
  fprintf(stream, "#Subr<%p=\"%s\">", get(obj, Subr,exec), tostring(get(obj, Subr,name)));
  return obj;
}

static oop print_Type(oop obj, FILE *stream, int cycle)
{
  fprintf(stream, "#Type(%s %ld", tostring(get(obj, Type,name)), get(obj, Type,size));
  oop fields= get(obj, Type,fields);
  if (!isNil(fields)) {
    for (;  isPair(fields);  fields= getTail(fields)) {
      fprintf(stream, " ");
      fprint(getHead(fields), stream);
    }
  }
  fprintf(stream, ")");
  return obj;
}

static oop print_Context(oop obj, FILE *stream, int cycle)
{
  fprintf(stream, "#Context(");
  oop func= get(obj, Context,function);
  if (notNil(func))
    fprint(func, stream);
  else
    fprintf(stream, "%p", obj);
  fprintf(stream, ")");
  return obj;
}

static oop print_type(oop obj, FILE *stream, int cycle)
{
  //#ifndef NDEBUG
  if (obj->_cycle) {
    fprintf(stream, "...");
    return obj;
  }
  obj->_cycle= cycle;
  //#endif
  fprintf(stream, "#%s(", tostring(get(getType(obj), Type,name)));
  oop fields;
  int i;
  for (fields= get(getType(obj), Type,fields), i= 0;  isPair(fields);  fields= getTail(fields), ++i) {
    oop field= getHead(fields);
    if (isSymbol(field)) {
      fprintf(stream, "%s: ", tostring(field));
      _print(getBody(obj)[i], stream, cycle);
    } else {
      _print(field, stream, cycle);
    }
    if (isPair(getTail(fields)))
      fprintf(stream, " ");
  }
  fprintf(stream, ")");
  return obj;
}
