/* eval.c -- compile and run s-expressions
** 
** Copyright (c) 2009 Ian Piumarta
** All Rights Reserved
** 
** Permission is hereby granted, free of charge, to any person obtaining a
** copy of this software and associated documentation files (the 'Software'),
** to deal in the Software without restriction, including without limitation
** the rights to use, copy, modify, merge, publish, distribute, and/or sell
** copies of the Software, and to permit persons to whom the Software is
** furnished to do so, provided that the above copyright notice(s) and this
** permission notice appear in all copies of the Software.  Inclusion of the
** the above copyright notice(s) and this permission notice in supporting
** documentation would be appreciated but is not required.
** 
** THE SOFTWARE IS PROVIDED 'AS IS'.  USE ENTIRELY AT YOUR OWN RISK.
** 
** Last edited: 2010-08-03 11:14:41 by piumarta on emilia-2.local
*/

static oop eval(oop expr, oop env);

static oop evprog(oop list, oop env)
{
  oop result= nil;
  for (;  isPair(list);  list= getTail(list)) {
    result= eval(getHead(list), env);
  }
  return result;
}

static oop apply_Pointer(oop func, oop args, int nArgs, oop env)
{
  long v[17];	/* fixme: count the args, then alloca() */
  int i;
  exec_t f= (exec_t)get(func, Pointer,address);
  oop r;
  for (i= 0;  nArgs-- && i < 15 && isPair(args);  args= cdr(args)) objectToPrimitive(car(args), v, &i, nil);
  if (nArgs >= 0) error("too many arguments in foreign call");
  switch (i) {
  case  0: r= f(); break;
  case  1: r= f(v[0]); break;
  case  2: r= f(v[0],v[1]); break;
  case  3: r= f(v[0],v[1],v[2]); break;
  case  4: r= f(v[0],v[1],v[2],v[3]); break;
  case  5: r= f(v[0],v[1],v[2],v[3],v[4]); break;
  case  6: r= f(v[0],v[1],v[2],v[3],v[4],v[5]); break;
  case  7: r= f(v[0],v[1],v[2],v[3],v[4],v[5],v[6]); break;
  case  8: r= f(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7]); break;
  case  9: r= f(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8]); break;
  case 10: r= f(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],v[9]); break;
  case 11: r= f(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],v[9],v[10]); break;
  case 12: r= f(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],v[9],v[10],v[11]); break;
  case 13: r= f(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],v[9],v[10],v[11],v[12]); break;
  case 14: r= f(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],v[9],v[10],v[11],v[12],v[13]); break;
  case 15: r= f(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],v[9],v[10],v[11],v[12],v[13],v[14]); break;
  default:
    return error("too many arguments in foreign function call");
  }
  return mkLong((long)r);
}

static oop apply(oop function, oop arguments, oop env)
{
  switch (getTypeIndex(function)) {
  case iPointer: {
    return apply_Pointer(function, arguments, list_length(arguments), env);
  }
  case iSubr: {
    return get(function, Subr,exec)(arguments, env);
  }
  case iExpr: {
    if (notNil(get(function, Expr,method))) {
      return exec(get(function, Expr,method), arguments, env);
    }
    oop params= get(function, Expr,params);
    oop body=   get(function, Expr,body);	GC_PROTECT(body);	// body might store into function.body
    oop env=    get(function, Expr,env);	GC_PROTECT(env);
    oop tmp=    nil;				GC_PROTECT(tmp);
    for (;  isPair(params);  params= getTail(params), arguments= cdr(arguments)) {
      tmp= cons(car(params), car(arguments));
      env= cons(tmp, env);
    }
    if (isSymbol(params)) {
      tmp= cons(params, arguments);
      env= cons(tmp, env);
    }
    tmp= evprog(body, env);			GC_UNPROTECT(body);
    return tmp;
  }
  case iSyntax: {
    GC_PROTECT(arguments);
    arguments= cons(nil, arguments);
    arguments= cons(nil, arguments);
    arguments= apply(get(function, Syntax,form), arguments, env);
    if (isNil(arguments)) return nil;
    return eval(arguments, env);
  }
  default:
    printsp(function);
    println(arguments);
    return error("cannot apply %s", tostring(get(getType(function), Type,name)));
  }
}

static oop evlist(oop list, oop env)
{
  if (!isPair(list)) return nil;
  oop tail= evlist(getTail(list), env);		GC_PROTECT(tail);
  oop head= eval(car(list), env);		GC_PROTECT(head);
  head= cons(head, tail);			GC_UNPROTECT(tail);
  return head;
}


#define arglist(ARGS) (ARGS)

static oop syntax_quote(oop arguments, oop env)
{
  arguments= arglist(arguments);
  return cadr(arguments);
}

static oop syntax_set(oop arguments, oop env)
{
  arguments= arglist(arguments);
  oop setter= cadr(arguments);
  if (isSymbol(setter)) {
    oop cell= assq(setter, env);
    if (isNil(cell)) return error("undefined: %s", tostring(setter));
    return set(cell, Pair,tail, evprog(cddr(arguments), env));
  }
  println(setter);
  return error("non-symbol identifier in set");
}

static oop syntax_define(oop arguments, oop env)
{
  arguments= arglist(arguments);
  oop name= cadr(arguments), program= cddr(arguments);
  if (!isSymbol(name)) error("non-symbol identifier in define");
  oop assoc= declare(name, env);
  oop value= evprog(program, env);
  if (isExpr(value) && isNil(get(value, Expr,name))) set(value, Expr,name, name);
  return set(assoc, Pair,tail, value);
}

static oop syntax_if(oop arguments, oop env)
{
  arguments= arglist(arguments);
  return notNil(eval(cadr(arguments), env)) ? eval(caddr(arguments), env) : evprog(cdddr(arguments), env);
}

static oop syntax_while(oop arguments, oop env)
{
  arguments= arglist(arguments);
  oop cond= cadr(arguments);
  oop body= cddr(arguments);
  oop result= nil;
  while (notNil(eval(cond, env))) result= evprog(body, env);
  return result;
}

static oop syntax_and(oop arguments, oop env)
{
  arguments= arglist(arguments);
  oop result= t;
  for (arguments= getTail(arguments);  isPair(arguments);  arguments= getTail(arguments))
    if (isNil(result= eval(car(arguments), env)))
      break;
  return result;
}

static oop syntax_or(oop arguments, oop env)
{
  arguments= arglist(arguments);
  oop result= nil;
  for (arguments= getTail(arguments);  isPair(arguments);  arguments= getTail(arguments))
    if (notNil(result= eval(car(arguments), env)))
      break;
  return result;
}

static oop syntax_let(oop arguments, oop env)
{
  arguments= arglist(arguments);
  oop lenv= env;				GC_PROTECT(lenv);
  oop tmp=  nil;				GC_PROTECT(tmp);
  oop bindings= cadr(arguments);
  for (;  isPair(bindings);  bindings= getTail(bindings)) {
    oop binding= getHead(bindings), name;
    if (isPair(binding) && isSymbol(name= getHead(binding))) {
      tmp= evprog(getTail(binding), env);
    } else if (isSymbol(name= binding)) {
      tmp= nil;
    } else {					GC_UNPROTECT(lenv);
      return error("syntax error in let");
    }
    tmp=  cons(name, tmp);
    lenv= cons(tmp, lenv);
  }
  tmp= evprog(cddr(arguments), lenv);		GC_UNPROTECT(lenv);
  return tmp;
}

static oop syntax_lambda(oop arguments, oop env)
{
  arguments= arglist(arguments);
  return mkExpr(arguments, cadr(arguments), cddr(arguments), nil, env);
}

#undef arglist


static oop eval(oop expr, oop env)
{
  {
#   undef static
    static char *cursors= "-\\|/";
    static int cursor= 0;
    static int count= 0;
    if (0 == (count++ & 0xfffff)) {
      if (0 == cursors[cursor]) cursor= 0;
      fprintf(stderr, "%c\010", cursors[cursor]);
      ++cursor;
    }
#  if (NONSTATIC)
#    define static
#  endif
  }
  if (optTrace & 1) { printf("EVAL "); println(expr); }
  switch (getTypeIndex(expr)) {
  case iSymbol: {
    oop cell= assq(expr, env);
    if (isNil(cell)) return error("undefined: %s", tostring(expr));
    return cdr(cell);
  }
  case iPair: {							GC_PROTECT(expr);
    backtrace= cons(expr, backtrace);
    oop func= car(expr);
    if (isSymbol(func)) {
      if (func == s_quote ) { expr= syntax_quote(expr, env);   goto done; }
      if (func == s_set	  ) { expr= syntax_set(expr, env);     goto done; }
      if (func == s_define) { expr= syntax_define(expr, env);  goto done; }
      if (func == s_if    ) { expr= syntax_if(expr, env);      goto done; }
      if (func == s_while ) { expr= syntax_while(expr, env);   goto done; }
      if (func == s_and	  ) { expr= syntax_and(expr, env);     goto done; }
      if (func == s_or	  ) { expr= syntax_or(expr, env);      goto done; }
      if (func == s_let	  ) { expr= syntax_let(expr, env);     goto done; }
      if (func == s_lambda) { expr= syntax_lambda(expr, env);  goto done; }
    }
    func= eval(func, env);					GC_PROTECT(func);
    if (iSyntax != getTypeIndex(func))
      expr= evlist(getTail(expr), env);
    expr= apply(func, expr, env);
   done:
    GC_UNPROTECT(expr);
    backtrace= cdr(backtrace);
    return expr;
  }
  default:
    return expr;
  }
}

oop compileFunction(oop function)
{
  if (optTrace & 16) { printf("COMPILE FUNCTION %p ", function);  printsp(function);  println(get(function, Expr,method)); }
  oop env= get(function, Expr,env);
  oop compiler= cdr(assq(s_compile_function, env));
  if (notNil(compiler)) {
    GC_PROTECT(function);
    oop args= cons(function, nil);		GC_PROTECT(args);
    args= apply(compiler, args, env);		GC_UNPROTECT(function);
    if (optTrace & 16) { printf("COMPILE FUNCTION %p => ", function);  println(args); }
    return set(function, Expr,method, args);
  }
  return nil;
}
