/* input.c -- lightweight input stream type
** 
** Copyright (c) 2009 Ian Piumarta
** All Rights Reserved
** 
** Permission is hereby granted, free of charge, to any person obtaining a
** copy of this software and associated documentation files (the 'Software'),
** to deal in the Software without restriction, including without limitation
** the rights to use, copy, modify, merge, publish, distribute, and/or sell
** copies of the Software, and to permit persons to whom the Software is
** furnished to do so, provided that the above copyright notice(s) and this
** permission notice appear in all copies of the Software.  Inclusion of the
** the above copyright notice(s) and this permission notice in supporting
** documentation would be appreciated but is not required.
** 
** THE SOFTWARE IS PROVIDED 'AS IS'.  USE ENTIRELY AT YOUR OWN RISK.
** 
** Last edited: 2010-08-03 11:14:36 by piumarta on emilia-2.local
*/

struct input
{
  FILE *file;
  char *buffer;
  char *string;
  int   ungotc;
};

typedef struct input INPUT;

static INPUT *ialloc(void)
{
  INPUT *in= malloc(sizeof(INPUT));
  in->file= 0;
  in->buffer= 0;
  in->string= 0;
  in->ungotc= 0;
  return in;
}

int igetc(INPUT *in)
{
  int c= in->ungotc;
  if (c) {
    in->ungotc= 0;
    return c;
  }
  else if (in->file) {
    return getc(in->file);
  }
  else if (in->string) {
    c= *in->string;
    if (c) {
      in->string++;
      return c;
    }
  }
  return -1;
}

int unigetc(int c, INPUT *in)
{
  return in->ungotc= c;
}

INPUT *iopen(char *s)
{
  FILE *f= fopen(s, "r");
  if (!f) return 0;
  INPUT *in= ialloc();
  int n= 0, m= 1024;
  in->buffer= malloc(m);
  for (;;) {
    if (n == m) in->buffer= realloc(in->buffer, m *= 2);
    int c= getc(f);
    if (c < 0) {
      in->buffer[n]= 0;
      fclose(f);
      in->string= in->buffer;
      break;
    }
    in->buffer[n++]= c;
  }
  return in;
}

INPUT *ifile(FILE *f)
{
  INPUT *in= ialloc();
  in->file= f;
  return in;
}

void iclose(INPUT *in)
{
  if (in->buffer) free(in->buffer);
  free(in);
}
