/* main.c -- definitions of builtin types and main program
** 
** Copyright (c) 2009 Ian Piumarta
** All Rights Reserved
** 
** Permission is hereby granted, free of charge, to any person obtaining a
** copy of this software and associated documentation files (the 'Software'),
** to deal in the Software without restriction, including without limitation
** the rights to use, copy, modify, merge, publish, distribute, and/or sell
** copies of the Software, and to permit persons to whom the Software is
** furnished to do so, provided that the above copyright notice(s) and this
** permission notice appear in all copies of the Software.  Inclusion of the
** the above copyright notice(s) and this permission notice in supporting
** documentation would be appreciated but is not required.
** 
** THE SOFTWARE IS PROVIDED 'AS IS'.  USE ENTIRELY AT YOUR OWN RISK.
** 
** Last edited: 2010-08-03 11:14:33 by piumarta on emilia-2.local
*/

#define OLD_ABI		0	// 1 => first argument is value of evaluated function being applied
#define USE_OT		0	// 1 => all objects indirect through { type, body } tuple
#define USE_NIL		0	// 1 => nil is a real object, not NULL
#define USE_TAG		1	// 1 => 30-bit longs can be tagged immediate values
#define USE_COMPILER	0	// 0 => use tree evaluator to bootstrap compiler; 1 => use built-in symcode compiler
#define NONSTATIC	0	// 1 => export all static symbols for profiling

#include "osdep.c"

#if !defined(GC)
# define GC	1
#endif

#if (GC == 0)				/* no GC */
# define GC_malloc(SIZE)	GC_realloc(0, (SIZE))
# define GC_malloc_atomic(SIZE)	GC_malloc(SIZE)
# define GC_size(o)		(((size_t *)(o))[-1])
  void *GC_realloc(void *o, size_t size)
  {
    if (size == 0) {
      free(o);
      return 0;
    }
    o= realloc((o ? &GC_size(o) : 0), sizeof(size_t) + size);
    if (!o)
      return 0;
    *((size_t *)o)= size;
    return &((size_t *)o)[1];
  }
# define GC_gcollect() /*empty*/
# define GC_register_finaliser(PTR, FINALISER, DATA) /*empty*/
#elif (GC == 1)				/* simple GC */
# define GC_API /* static */
# include "gc.c"
#elif (GC == 2)				/* Boehm GC */
# include <sys/types.h>
  extern void   GC_init(void);
  extern void  *GC_malloc(size_t);
  extern void  *GC_malloc_atomic(size_t);
  extern void  *GC_realloc(void *, size_t);
  extern size_t GC_size(void *);
  extern void   GC_gcollect(void);
  extern void  *GC_realloc(void *, size_t);
  extern size_t GC_size(void *);
  extern void   GC_gcollect(void);
  typedef void (*GC_finaliser_t)(void *ptr, void *data);
  struct gcfinalise
  {
    GC_finaliser_t finaliser;
    void *data;
    struct gcfinalise *next;
  };
  static void GC_multi_finalise(void *obj, void *client_data)
  {
    struct gcfinalise *p= client_data;
    while (p) {
      (*p->finaliser)(obj, p->data);
      p= p->next;
    }
  }
  void GC_register_finaliser(void *ptr, GC_finaliser_t finaliser, void *data)
  {
    extern void  GC_register_finalizer(void *obj, GC_finaliser_t proc, void *cd, GC_finaliser_t *oproc, void **ocd);
    GC_finaliser_t tmp;
    struct gcfinalise *f= GC_malloc(sizeof(*f));
    f->finaliser= finaliser;
    f->data= data;
    f->next= 0;
    /* Boehm GC only accepts a single finaliser, so we have to implement our own linked list. */
    GC_register_finalizer(ptr, GC_multi_finalise, f, &tmp, (void **)(void *)&f->next);
  }
#endif

#if (GC != 1)
# define GC_add_root(X)
# define GC_PROTECT(X)
# define GC_UNPROTECT(X)
# define GC_check(X)		(X)
# define GC_mark(X)
# define GC_mark_leaf(X)
#endif

#if (NONSTATIC)
# define static
#endif

struct Object;

typedef struct Object *oop;

typedef oop  (*exec_t)();

typedef unsigned char byte;
typedef unsigned long word;

static oop fprint(oop arg, FILE *stream);
static oop fprintsp(oop arg, FILE *stream);
static oop fprintln(oop arg, FILE *stream);
static oop print(oop arg);
static oop printsp(oop arg);
static oop println(oop arg);

#define typesDo(_) _(Undefined) _(Tag) _(Long) _(Float) _(Pointer) _(Pair) _(String) _(Symbol) _(Array) _(ByteArray) _(WordArray) _(Syntax) _(Expr) _(Subr) _(Context)

typedef enum {
#define i(TYPE) , i##TYPE
  iType= 0
  typesDo(i)
#undef i
} typeIndex_t;

typedef void (*mark_t)(oop obj);
typedef oop  (*print_t)(oop obj, FILE *stream, int cycle);

#define t(TYPE) static oop t##TYPE;
t(Type)
typesDo(t)
#undef t

typedef struct {
}		 Undefined;

typedef struct {
}		 Tag;

typedef struct {
  long	 value;
}		 Long;

typedef struct {
  double value;
}		 Float;

typedef struct {
  void	*address;
  oop	 name;			/* Symbol */
  oop	 type;			/* list */
}		 Pointer;

typedef struct {
  oop	 head;
  oop	 tail;
}		 Pair;

typedef struct {
  oop	 size;
  char	*elements;
}		 String;

typedef struct {
  oop	 string;
  int	 opcode;
}		 Symbol;

typedef struct {
  oop	 size;
  oop	 undefined;
  oop	*elements;
}		 Array;

typedef struct {
  oop	 size;
  byte	*elements;
}		 ByteArray;

typedef struct {
  oop	 size;
  word	*elements;
}		 WordArray;

typedef struct {
  oop	 form;			/* Subr, Expr */
  oop	 compiler;		/* Subr, Expr */
}		 Syntax;

typedef struct {
  oop	 expr;
  oop	 params;
  oop	 body;
  oop	 method;
  oop	 env;
  oop	 name;
}		 Expr;

typedef struct {
  exec_t exec;
  oop	 name;			/* Symbol */
  long	 count;
}		 Subr;

typedef struct {
  typeIndex_t	 index;
  oop		 name;
  oop		 fields;
  long		 size;
  mark_t	 mark;
  print_t	 print;
}		 Type;

typedef struct {
  oop		 pc;
  oop		 env;
  oop		 caller;
  oop		 callee;
  oop		 function;
  int		 ip;
  oop		*sp;
  oop		*stackBase;
  oop		*stackLimit;
  oop		 kernel;
}		 Context;

/* Indices in the Context.kernel array: */
enum {
  CONTEXT_KERNEL_PARENT = 0, /* The kernel of our kernel. */
  CONTEXT_KERNEL_ENTRY	= 1, /* The entry point to our kernel. */
  CONTEXT_KERNEL_BUDGET = 2, /* How many ticks to ignore before calling (kernel-entry user). */
  CONTEXT_KERNEL_END
};

typedef union
{
#define u(TYPE) TYPE _##TYPE;
  u(Type)
  typesDo(u)
#undef u
} body_t;

struct Object
{
  short		 _type;			// xxx
  oop		 properties;
  int		 _cycle;
#ifndef NDEBUG
  const char	*_file;
  int		 _line;
  const char	*_func;
#endif
#if USE_OT
  body_t	*_body;
#else
  body_t	 _body;
#endif
};

static oop types= 0;

#if USE_NIL
# if USE_TAG
#   define getTypeIndex(OBJECT)		(((long)(OBJECT)&1) ? iLong : ((OBJECT)->_type              ))
#   define getProperties(OBJECT)	(((long)(OBJECT)&1) ? nil   : ((OBJECT)->properties         ))
#   define setProperties(OBJECT, PROPS)	(((long)(OBJECT)&1) ? nil   : ((OBJECT)->properties= (PROPS)))
# else
#   define getTypeIndex(OBJECT)		((OBJECT)->_type              )
#   define getProperties(OBJECT)	((OBJECT)->properties         )
#   define setProperties(OBJECT, PROPS)	((OBJECT)->properties= (PROPS))
# endif
#else
# if USE_TAG
#   define getTypeIndex(OBJECT)		((OBJECT) ? (((long)(OBJECT)&1) ? iLong : ((OBJECT)->_type              )) : iUndefined)
#   define getProperties(OBJECT)	((OBJECT) ? (((long)(OBJECT)&1) ? nil   : ((OBJECT)->properties         )) : nil       )
#   define setProperties(OBJECT, PROPS)	((OBJECT) ? (((long)(OBJECT)&1) ? nil   : ((OBJECT)->properties= (PROPS))) : nil       )
# else
#   define getTypeIndex(OBJECT)		((OBJECT) ? ((OBJECT)->_type     	  ) : iUndefined)
#   define getProperties(OBJECT)	((OBJECT) ? ((OBJECT)->properties	  ) : nil       )
#   define setProperties(OBJECT, PROPS)	((OBJECT) ? ((OBJECT)->properties= (PROPS)) : nil       )
# endif
#endif

#if USE_OT
# define getBody(OBJECT)		((oop *)((OBJECT)->_body))
# define _get(OBJECT, FIELD)		(gc_check(OBJECT)->_body->_##FIELD)
# define _set(OBJECT, FIELD, VALUE)	(gc_check(OBJECT)->_body->_##FIELD= (VALUE))
#else
# define getBody(OBJECT)		((oop *)&((OBJECT)->_body))
# define _get(OBJECT, FIELD)		(gc_check(OBJECT)->_body._##FIELD)
# define _set(OBJECT, FIELD, VALUE)	(gc_check(OBJECT)->_body._##FIELD= (VALUE))
#endif

#if defined(DEBUG)
# define get(OBJECT, TYPE,FIELD)	({ oop _O= (OBJECT);  assert(getTypeIndex(_O) == i##TYPE);  _get(_O, TYPE.FIELD); })
# define set(OBJECT, TYPE,FIELD, VALUE)	({ oop _O= (OBJECT);  assert(getTypeIndex(_O) == i##TYPE);  _set(_O, TYPE.FIELD, VALUE); })
#else
# define get(OBJECT, TYPE,FIELD)	_get(OBJECT, TYPE.FIELD)
# define set(OBJECT, TYPE,FIELD, VALUE)	_set(OBJECT, TYPE.FIELD, VALUE)
#endif

#if USE_NIL
  oop nil;
  static int isNil(oop obj)	{ return nil == obj; }
  static int notNil(oop obj)	{ return nil != obj; }
#else
# define nil 0
# define isNil(OBJ)	(!(OBJ))
# define notNil(OBJ)	(  OBJ )
#endif

#define getType(OBJECT)		(get(types, Array,elements)[getTypeIndex(OBJECT)])

#if defined(NDEBUG)
# define gc_check(O) (O)
#else
  static oop gc_check(oop o)
  {
    if (!GC_check(o)) {
      fprintf(stderr, "accessible dead object %p:%i ", o, (int)getTypeIndex(o));
      fprintln(o, stderr);
      abort();
    }
    return o;
  }
#endif

#define _(TYPE) static inline int is##TYPE(oop obj)	{ return i##TYPE == getTypeIndex(obj); }
_(Type)
typesDo(_)
#undef _

#if USE_TAG
  static inline long getLong(oop obj)	{ return ((long)obj & 1) ? ((long)obj >> 1) : get(obj, Long,value); }
# else
# define getLong(O)	get((O), Long,value)
#endif

#define getFloat(O)	get((O), Float,value)
#define getHead(O)	get((O), Pair,head)
#define getTail(O)	get((O), Pair,tail)

static oop t;
static oop symbols;

static oop s_globals;

static oop s_quote;
static oop s_set;
static oop s_define;
static oop s_if;
static oop s_while;
static oop s_and;
static oop s_or;
static oop s_let;
static oop s_lambda;

static oop s_eval;
static oop s_quasiquote;
static oop s_unquote;
static oop s_unquote_splicing;

static oop s_compile_function;

static oop s_oopAt;
static oop s_ptrAt;

static oop s_function;
static oop s_pointer;
static oop s_oop;
static oop s_void;
static oop s_char;
static oop s_short;
static oop s_int;
static oop s_long;
static oop s_float;
static oop s_double;

static oop s_argv;
static oop s_args;
static oop s_input_stream;

static oop t_pointer_void;

#define isT(O)	(t == (O))

static void printBacktrace(void);

static oop warning(char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  fprintf(stderr, "WARNING: ");
  vfprintf(stderr, fmt, ap);
  fprintf(stderr, "\n");
  va_end(ap);
  return nil;
}

static oop error(char *fmt, ...)
{
  printBacktrace();
  va_list ap;
  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  fprintf(stderr, "\n");
  va_end(ap);
  exit(1);
  return nil;
}

#define cons(A, B)		mkPair((A), (B))

static oop car(oop list)	{ return isPair(list) ? get(list, Pair,head) : nil; }
static oop cdr(oop list)	{ return isPair(list) ? get(list, Pair,tail) : nil; }

static oop caar(oop list)	{ return car(car(list)); }
static oop cadr(oop list)	{ return car(cdr(list)); }
static oop cdar(oop list)	{ return cdr(car(list)); }
static oop cddr(oop list)	{ return cdr(cdr(list)); }
static oop cadar(oop list)	{ return car(cdr(car(list))); }
static oop caddr(oop list)	{ return car(cdr(cdr(list))); }
static oop cdddr(oop list)	{ return cdr(cdr(cdr(list))); }
static oop cadddr(oop list)	{ return car(cdr(cdr(cdr(list)))); }
//static oop cddddr(oop list)	{ return cdr(cdr(cdr(cdr(list)))); }
static oop caddddr(oop list)	{ return car(cdr(cdr(cdr(cdr(list))))); }
static oop cadddddr(oop list)	{ return car(cdr(cdr(cdr(cdr(cdr(list)))))); }

static oop rplaca(oop list, oop value)	{ return set(list, Pair,head, value); }
static oop rplacd(oop list, oop value)	{ return set(list, Pair,tail, value); }

static oop assoc(oop key, oop list);
static oop array_append(oop array, oop element);

char *optBootFile= "boot.k";
long  optInteractive= 1;
long  optVerboseStats= 0;
long  optVerboseRead= 0;
long  optVerbosePrint= 0;
long  optVerboseCompile= 0;
long  optVerboseExec= 0;
long  optCompile= 0;
long  optTrace= 0;

#define arithmeticDo(_)  _(add) _(sub) _(mul) _(div) _(mod) _(asl) _(asr) _(and) _(or) _(xor) _(lt) _(le) _(ge) _(gt) _(eq) _(neq)

#define specialsDo(_)  _(cons) _(car) _(cdr) _(setcar) _(setcdr) _(eql) _(neql) arithmeticDo(_)

#define insnsDo(_) _(push) _(lookup) _(assign) _(dup) _(pop) _(drop) _(bind) _(arg) _(rest) _(popenv) _(load) _(name) _(store) _(bf) _(bt) _(br) _(lambda) _(tick) _(xapply) _(return) _(xreturn) _(xdrop) _(context) _(resume) _(addr) specialsDo(_)

#define i(insn) static oop s_##insn;
insnsDo(i)
#undef i

#define i(insn) static oop subr_##insn;
arithmeticDo(i)
#undef i

#include "allocate.c"
#include "intrinsic.c"
#include "exec2.c"
#if USE_COMPILER
# include "compile.c"
#else
# include "eval.c"
#endif
#include "input.c"
#include "reader.c"
#include "debug.c"
#include "types.c"
#include "primitive.c"

static void trap(char *what)
{
  printf("\n%s\n", what);
  printBacktrace();
  exit(1);
}

static void interrupt(int signo)	{ trap("Interrupt"); }
static void segFault(int signo)		{ trap("Segmentation fault"); }
static void busError(int signo)		{ trap("Bus error"); }
static void abortTrap(int signo)	{ trap("Abort trap"); }

static oop repl(INPUT *stream, oop env)
{
						GC_PROTECT(env);
  oop o= assq(s_input_stream, env);		GC_PROTECT(o);
  if (isNil(o)) o= cons(s_input_stream, o);
  rplacd(o, mkPointer(stream, s_input_stream, nil));
  oop result= nil;
  for (;;) {
    if (stream->file == stdin) {
      printf(".");
      fflush(stdout);
    }
    if (readerEOF == (o= readObject(stream, env))) break;
    if (optVerboseRead) {
      print(o);
      if (optVerbosePrint)
	printf("\n ==> ");
      else
	printf("\n");
    }
    oop reval= assq(s_eval, env);
    if (isNil(reval))
      result= eval(o, env);
    else {
      o= cons(o, nil);
      rplacd(o, cons(env, nil));
      result= apply(cdr(reval), o, env);
    }
    if (optVerbosePrint) println(result);
  }
						GC_UNPROTECT(env);
  return result;
}

#if !defined(OFW)
static long mktol(char *mk)
{
  char *endptr;
  long l= strtol(mk, &endptr, 0);
  if (endptr == mk) error("invalid number: %s", mk);
  if ('g' == *endptr || 'G' == *endptr) l *= 1024*1024*1024;
  if ('m' == *endptr || 'M' == *endptr) l *= 1024*1024;
  if ('k' == *endptr || 'K' == *endptr) l *= 1024;
  return l;
}
#endif

static void boot(char *progname, oop env)
{
  char base[PATH_MAX], *namep= progname, *out= base, *delim= 0, *baseurl= "";
  INPUT *in= 0;
  if (optBootFile && (in= iopen(optBootFile))) {
    delim= strrchr(optBootFile, '/');
    if (delim) {
      baseurl= optBootFile;
      delim[1]= '\0';
    }
  }
  else {
    // 1: look for a path to the executable file
    for (;  *namep;  *out++= *namep++)
      if ('\\' == *namep || '/' == *namep) delim= out;
#  if !defined(OFW)
    if (!delim) {
      // 2: look for the executable in PATH
      char *paths= getenv("PATH");
      for (namep= strtok(paths, ":");  namep;  namep= strtok(0, ":")) {
	sprintf(base, "%s/%s", namep, progname);
	if (!access(base, R_OK | X_OK)) {	// this is NOT guaranteed to find the right file
	  delim= strrchr(base, '/');
	  break;
	}
      }
    }
#  endif
    if (delim) {
      // 3: make a dirname from the location of the delimiter
      delim[1]= '\0';
      // 4: if dirname ends in /bin then change it to /lib/cola
      if ((delim - base > 3) && !strcmp(delim - 4, "/bin/")) strcpy(delim - 4, "/lib/cola/");
      baseurl= &base[0];
    }
  }
  oop name= mkSymbol("*baseurl*");
  declare(name, env);
  define(name, mkString(baseurl), env);
  if (!optBootFile) return;
  if (!in) {
    sprintf(base, "%s%s", baseurl, optBootFile);
    in= iopen(base);
    if (!in) error("%s: no such file or directory", base);
  }
  if (!in) error("%s: no such file or directory", optBootFile);
  repl(in, env);
  iclose(in);
  optBootFile= 0;
}

int main(int argc, char *argv[])
{
  char *progname= argv[0];
  //printf(";booting\n");
  //setbuf(stdout, 0);
  setbuf(stderr, 0);

#if (GC == 1)
  GC_mark_function= markFunction;
#elif (GC == 2)
  GC_init();
#endif

#if USE_NIL
  nil= empty(iUndefined);				GC_add_root(&nil);
#endif
  types= nil;						GC_add_root(&types);
  backtrace= nil;					GC_add_root(&backtrace);
  symbols= nil;						GC_add_root(&symbols);

  makeTypes();

  t=			mkSymbol("t");			/* protected via symbols */
  s_globals=		mkSymbol("*globals*");
  s_quote=		mkSymbol("quote");
  s_set=		mkSymbol("set");
  s_define=		mkSymbol("define");
  s_if=			mkSymbol("if");
  s_while=		mkSymbol("while");
  s_and=		mkSymbol("and");
  s_or=			mkSymbol("or");
  s_let=		mkSymbol("let");
  s_lambda=		mkSymbol("lambda");
  s_quasiquote=		mkSymbol("quasiquote");
  s_unquote=		mkSymbol("unquote");
  s_unquote_splicing=	mkSymbol("unquote-splicing");
  s_eval=		mkSymbol("eval");
  s_compile_function=	mkSymbol("compile-function");
  s_oopAt=		mkSymbol("oop@");
  s_ptrAt=		mkSymbol("pointer@");
  s_function=		mkSymbol("function");
  s_pointer=		mkSymbol("*");
  s_void=		mkSymbol("void");
  s_char=		mkSymbol("char");
  s_short=		mkSymbol("short");
  s_int=		mkSymbol("int");
  s_long=		mkSymbol("long");
  s_float=		mkSymbol("float");
  s_double=		mkSymbol("double");
  s_argv=		mkSymbol("*argv*");
  s_args=		mkSymbol("*args*");
  s_input_stream=	mkSymbol("*input-stream*");

  t_pointer_void=	cons(s_void, nil);			GC_add_root(&t_pointer_void);
  t_pointer_void=	cons(s_pointer, t_pointer_void);	GC_add_root(&t_pointer_void);

  int n= 0;

#define i(insn) s_##insn = mkSymbol(#insn);  set(s_##insn, Symbol,opcode, ++n);
insnsDo(i)
#undef i

  initTypes();

#if (GC == 1)
  GC_pre_mark_function= mark_threadSpecificLocation;
#endif
  threadSpecific=       mkArray(CONTEXT_KERNEL_END);
  {
    oop *elements= get(threadSpecific, Array,elements);
    elements[CONTEXT_KERNEL_BUDGET]= mkLong(1); /* Never fire the entry point. */
  }

  oop env= nil;						GC_PROTECT(env);
  oop sym= nil;						GC_PROTECT(sym);
  oop sub= nil;						GC_PROTECT(sub);

#define prim(SYM, ID) {				\
    sym= mkSymbol(SYM);				\
    sub= mkSubr(Subr_##ID, sym);		\
    sym= mkPair(sym, sub);			\
    env= mkPair(sym, env);			\
  }

  prim("environment",		environment);
  prim("error",			error);
  prim("trace",			trace);
  prim("properties",		properties);
  prim("properties-at",		properties_at);
  prim("set-properties-at",	set_properties_at);
  prim("read",			read);
  prim("apply",			apply);
  prim("eval",			eval);
  prim("exec",			exec);
  prim("dump",			dump);
  prim("print",			print);
  prim("long",			long);
  prim("long?",			longP);
  prim("float",			float);
  prim("float?",		floatP);
  prim("finalise",		finalise);
  prim("pointer",		pointer);
  prim("pointer?",		pointerP);
  prim("pointer-address",	pointer_address);
  prim("set-pointer-address",	set_pointer_address);
  prim("syntax",		syntax);
  prim("syntax?",		syntaxP);
  prim("syntax-function",	syntax_function);
  prim("set-syntax-function",	set_syntax_function);
  prim("new-type",		new_type);
  prim("new",			new);
  prim("typeof",		typeof);
  prim("type-index",		type_index);
  prim("type-name",		type_name);
  prim("type-fields",		type_fields);
  prim("string",		string);
  prim("string?",		stringP);
  prim("string-length",		string_length);
  prim("string-at",		string_at);
  prim("set-string-at",		set_string_at);
  prim("symbol",		symbol);
  prim("symbol?",		symbolP);
  prim("symbol-opcode",		symbol_opcode);
  prim("array",			array);
  prim("array-length",		array_length);
  prim("array-swap",		array_swap);
  prim("array-at",		array_at);
  prim("array-at-typeof",	array_at_typeof);
  prim("set-array-at",		set_array_at);
  prim("byte-array",		byte_array);
  prim("byte-array-length",	byte_array_length);
  prim("byte-array-at",		byte_array_at);
  prim("set-byte-array-at",	set_byte_array_at);
  prim("word-array",		word_array);
  prim("word-array-length",	word_array_length);
  prim("word-array-at",		word_array_at);
  prim("set-word-array-at",	set_word_array_at);
  prim("dlopen",		dlopen);
  prim("extern",		extern);
  prim("extern?",		externP);
  prim("primitive?",		primitiveP);
  prim("function",		function);
  prim("function?",		functionP);
  prim("function-parameters",	function_parameters);
  prim("function-body",		function_body);
  prim("function-method",	function_method);
  prim("set-function-method",	set_function_method);
  prim("function-environment",	function_environment);
  prim("context",		context);
  prim("context?",		contextP);
  prim("context-pc",		context_pc);
  prim("context-stack",		context_stack);
  prim("context-env",		context_env);
  prim("set-context-env",	set_context_env);
  prim("context-caller",	context_caller);
  prim("context-ip",		context_ip);
  prim("context-kernel",	context_kernel);
  prim("set-context-kernel",	set_context_kernel);
  prim("cons",			cons);
  prim("pair?",			pairP);
  prim("list-length",		list_length);
  prim("car",			car);
  prim("set-car",		set_car);
  prim("cdr",			cdr);
  prim("set-cdr",		set_cdr);
  prim("assoc",			assoc);
  prim("assq",			assq);
  prim("nssq",			nssq);
  prim("concat",		concat);
  prim("long@",			longAt);
  prim("set-long@",		set_longAt);
  prim("set-if-long@",		setIf_longAt);
  prim("oop@",			oopAt);
  prim("set-oop@",		set_oopAt);
  prim("set-if-oop@",		setIf_oopAt);
  prim("not",			not);
  prim("=",			equal);
  prim("!=",			nequal);

  prim("inb",			inb);
  prim("outb",			outb);

#undef prim
#define prim(SYM, ID) {				\
    sym= mkSymbol(SYM);				\
    sub= mkSubr(Subr_##ID, sym);		\
    sym= mkPair(sym, sub);			\
    env= mkPair(sym, env);			\
    subr_##ID=sub;				\
    GC_add_root(&subr_##ID);			\
  }

  prim("+",			add);
  prim("-",			sub);
  prim("*",			mul);
  prim("/",			div);
  prim("%",			mod);
  prim("<<",			asl);
  prim(">>",			asr);
  prim("&",			and);
  prim("|",			or);
  prim("^",			xor);
  prim("<",			lt);
  prim("<=",			le);
  prim("==",			eq);
  prim("!==",			neq);
  prim(">=",			ge);
  prim(">",			gt);

#undef prim
							GC_UNPROTECT(sub);
							GC_UNPROTECT(sym);
  {
    oop arga= mkArray(0);				GC_PROTECT(arga);
    env= mkPair(nil, env);
    rplaca(env, mkPair(s_argv, arga));
    char **argp;
    int    argn;
    for (argn= argc,  argp= argv;  argn;  ++argp,  --argn)
      array_append(arga, mkString(*argp));
							GC_UNPROTECT(arga);
  }

  oop args= nil;					GC_add_root(&args);

  {
    int argn;
    for (argn= argc;  --argn;) {
      args= mkPair(nil, args);
      rplaca(args, mkString(argv[argn]));
    }
    env= mkPair(nil, env);
    args= rplaca(env, mkPair(s_args, args));
  }

  env= mkPair(nil, env);
  rplaca(env, mkPair(s_input_stream, nil));

  env= makeNamespace(env, s_globals);

#if !defined(OFW)
  signal(SIGINT,  interrupt);
  signal(SIGSEGV, segFault);
# if !defined(WIN32)
  signal(SIGBUS,  busError);
# endif
  signal(SIGABRT, abortTrap);
#endif

  for (;;) {
    if (isNil(cdr(args))) break;
    char *opt= strdup(get(cadr(args), String,elements));
    rplacd(args, cddr(args));
    char *arg= isPair(cdr(args)) ? get(cadr(args), String,elements) : 0;
    if      (!strcmp("-n",   opt)) optBootFile= 0;
    else if (!strcmp("-vr",  opt)) optVerboseRead= 1;
    else if (!strcmp("-vp",  opt)) optVerbosePrint= 1;
#  if !defined(OFW)
    else if (!strcmp("-vg",  opt)) optVerboseStats |= 1;
    else if (!strcmp("-vi",  opt)) optVerboseStats |= 2;
    else if (!strcmp("-vf",  opt)) optVerboseStats |= 4;
    else if (!strcmp("-vm",  opt)) optVerboseStats |= 7;
#  endif
    else if (!strcmp("-vc",  opt)) optVerboseCompile= 1;
    else if (!strcmp("-vx",  opt)) optVerboseExec= 1;
    else if (!strcmp("-v",   opt)) optVerboseRead= optVerbosePrint= 1;
    else if (!strcmp("-O",   opt)) optCompile= 1;
#  if (GC == 1) && !defined(OFW)
    else if (!strcmp("-gca", opt) && arg) { gcCount= gcAllocs= mktol((--argc, opt));  		rplacd(args, cddr(args)); }
    else if (!strcmp("-gcm", opt) && arg) { GC_gcollect();  gcMemory= mktol((--argc, opt));	rplacd(args, cddr(args)); }
    else if (!strcmp("-gcq", opt) && arg) { gcQuantum= mktol((--argc, opt));			rplacd(args, cddr(args)); }
#  endif
    else if (!strcmp("-vt",  opt) && arg) { optTrace= atoi((--argc, opt));			rplacd(args, cddr(args)); }
    else if (!strcmp("-b",   opt) && arg) { optBootFile= (--argc, opt);				rplacd(args, cddr(args)); }
    else if (!strcmp("-",    opt)) {
      int savVerbosePrint= optVerbosePrint;
      optInteractive= 1;
      boot(progname, env);
      INPUT *in= ifile(stdin);
      optVerbosePrint= 1;
      repl(in, env);
      iclose(in);
      optInteractive= 0;
      optVerbosePrint= savVerbosePrint;
    }
    else {
      optInteractive= 0;
      boot(progname, env);
      //printf("; loading: %s\n", opt);
      fflush(stdout);
      INPUT *stream= iopen(opt);
#    if defined(OFW)
      int n= 0;
      while (!stream && (n++ < 100)) {
	sleep(1);
	//fprintf(stderr, "; trying %d\r", n);
	fflush(stderr);
	stream= fopen(opt, "r");
	if (stream) fprintf(stderr, "\n");
      }
#    endif
      if (!stream)
	error("%s: no such file or directory", opt);
      else {
	repl(stream, env);
	iclose(stream);
      }
    }
  }

  if (optInteractive) {
    printf("Welcome to Chains of Meaning bootstrap 2.99\n");
    boot(progname, env);
    INPUT *in= ifile(stdin);
    optVerbosePrint= 1;
    repl(in, env);
  }

  if (optInteractive || optVerboseStats) {
    GC_gcollect();
#  if (GC == 1)
    if (!optVerboseStats) {
      fprintf(stdout, "\n[%ld bytes %ld objects %ld collections]\n", (long)GC_count_bytes(), (long)GC_count_objects(), GC_collections);
    }
    if (optVerboseStats & 1) {
      fprintf(stdout, "\n---------------- memory\n");
      fprintf(stdout, "%-7ld bytes\n",       (long)GC_count_bytes());
      fprintf(stdout, "%-7ld objects\n",     (long)GC_count_objects());
      fprintf(stdout, "%-7ld collections\n", (long)GC_collections);
    }
#  endif
#  if !defined(NDEBUG)
    if (optVerboseStats & 2) {
      fprintf(stdout, "\n---------------- instructions\n");
#    define i(insn) printf("%-7ld insn %s\n", insn##_count, #insn);
      insnsDo(i);
#    undef i
    }
    if (optVerboseStats & 4) {
      oop globals= localNamespace(env);
      fprintf(stdout, "\n---------------- primitives\n");
      while (notNil(globals)) {
	oop value= cdar(globals);
	if (isSubr(value)) {
	  printf("%-7ld prim ", get(value, Subr,count));
	  println(get(value, Subr,name));
	}
	globals= cdr(globals);
      }
    }
#  endif
  }

  return 0;

#undef isLong
  (void)isLong;
#if (GC == 1)
  (void)GC_delete_root;
  (void)GC_check;
#endif
  (void)mktol;
  (void)markFunction;
  (void)mark_threadSpecificLocation;
}
