# -*- fundamental -*-

digit8		= [01234567]
digit10		= [0123456789]
digit16		= [0123456789abcdefABCDEF]
alphabetic	= [ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_]
alphanumeric	= alphabetic | digit10 | "-"
blank		= [ \t\n\r]
eol		= "\r\n" | "\n" | "\r"
eof		= !.
comment		= "#" ( !eol . )*
_		= ( blank | comment )*

identifier	= ( alphabetic alphanumeric* ) @ $ $$ :id _ -> :id
field		= "." identifier :id _ -> :id

symbol0		= [-!$%&*+,./:<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ\\^_abcdefghijklmnopqrstuvwxyz|~]
symbol1		= [-!$%&*+,./:<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ\\^_abcdefghijklmnopqrstuvwxyz|~0123456789]
symbol		= ( symbol0 symbol1* ) @ $ $$ :s _ -> :s

number		= "0x" digit16+ $ $#16 :n _ -> :n
		|      digit10+ $ $#10 :n _ -> :n

equal		= "="  	_
equalplus	= "=+" 	_
plusequal	= "+=" 	_
colontpl	= ":::" _
colondbl	= "::" 	_
colon		= ":"  	_
semicolon	= ";"  	_
bar		= "|"  	_
ampersand	= "&"  	_
pling		= "!"  	_
question	= "?"  	_
star		= "*"  	_
plus		= "+"  	_
minus		= "-"  	_
quote		= "'"  	_
lparen		= "("  	_
rparen		= ")"  	_
tick		= "`"   _
ticklbrace	= "`{"  _
lbrace		= "{"  	_
rbrace		= "}"  	_
langle		= "<"  	_
rangle		= ">"  	_
dot		= "."  	_
dollardbl	= "$$" 	_
dollardot	= "$." 	_
dollarhash	= "$#" 	_
dollar		= "$"  	_
comma		= ","  	_
commaat		= ",@"  _
at		= "@"  	_
rarrow		= "->" 	_

escapedChar	= "\\"  ( "e"  -> 27
			| "a"  ->  7
			| "b"  ->  8
			| "f"  -> 12
			| "n"  -> 10
			| "r"  -> 13
			| "t"  ->  9
			| "v"  -> 11
			| "\"" -> 34
			| "'"  -> 39
			| "\\" -> 92
			| ( digit8 digit8 digit8 ) @ $ $#8
			)

char		= escapedChar | .

string		= "\"" ( !"\"" char )* $ :s "\"" _			-> :s

class		= "[" ( !"]" char )* $ :s "]" _				-> :s

argument	= identifier :i						-> (quote :i)
		| colon identifier
		| number
		| string

rule		= langle colon		   identifier:id		rangle	-> (match-object :id)
		| langle                   identifier:id argument*:args rangle	-> (match-rule      :id :: args)
		| langle identifier:ns dot identifier:id argument*:args rangle	-> (match-rule2 :ns :id :: args)
		|			   identifier:id !(field* ("="|"+="))	-> (match-rule      :id)

sexpr		= number
		| quote   sexpr:e					-> (quote      	     :e)

		| tick    sexpr:e					-> (quasiquote 	     :e)
		| commaat sexpr:e					-> (unquote-splicing :e)
		| comma   sexpr:e					-> (unquote	     :e)
		| string
		| symbol
		| lparen sexpr*:e rparen				-> :e

convname	= ("{" identifier:i "}") | identifier:i			-> :i

conversion	= "\\#" convname:i					-> (match-format-d :i)
		| "\\$" convname:i					-> (match-format-s :i)
		| char:x						-> (match-format-c :x)

format		= "\"" ( !"\"" conversion )*:c "\"" _			-> (match-format ::c)

action		= ticklbrace sexpr*:e rbrace				->  :e
		| tick format:e						-> (:e)
		| tick sexpr:e						-> (:e)

primary		= rule
		| colon identifier:id					-> (match-store :id)
 		| quote symbol:i					-> (match-object (quote :i))
		| quote lparen expression:e rparen			-> (match-struct :e)
		| lbrace expression:e rbrace				-> :e
		| lparen expression:e rparen				-> :e
		| action:e						-> (match-action :: e)
		| number:e						-> (match-object :e)
		| string:s						-> (match-string :s)
		| class:s						-> (match-class :s)
		| dot							-> (match-any)

suffix		= primary:p ( question					-> (match-zero-one  :p)
			    | star					-> (match-zero-more :p)
			    | plus					-> (match-one-more  :p)
			    |						-> :p
			    )

predicate	= tick symbol:s						-> (match-test-predicate :s)
		| tick sexpr:s						-> (match-test-expression :s)
		|      prefix

prefix		= ampersand predicate:p					-> (match-pred :p)
		| pling     predicate:p					-> (match-not  :p)
		|           suffix

varref		= identifier:var ( field:fld _				-> (match-field :var :fld)
				 |					-> :var
				 )

element		= number:x						-> (match-append :x)
		| colontpl varref:var					-> (match-append-all-all :var)
		| colondbl varref:var					-> (match-append-all :var)
		| colon    varref:var					-> (match-append :var)
		| list:l						-> (match-append-result :l)
		| symbol:id						-> (match-append (quote :id))
		| string:s						-> (match-append :s)

list		= lparen element*:l rparen				-> (match-group-1 :: l)

result		= rarrow ( colon varref:x				-> (match-result :x)
			 | number:x					-> (match-result :x)
			 | tick format
			 | action:x					-> (match-value  ::x)
			 | symbol:x					-> (match-result (quote :x))
			 | string:x					-> (match-result :x)
			 | list
			 )

atom		= ( result | prefix ) :expr
			( dollardbl					-> (match-intern    :expr) :expr
			| dollarhash number:n				-> (match-number :n :expr) :expr
			| dollardot					-> (match-float     :expr) :expr
			| dollar					-> (match-text      :expr) :expr
			| at						-> (match-input     :expr) :expr
			)*						-> :expr

sequence	= atom*:s						-> (match-all      :: s)
expression	= sequence:s ( bar sequence )*:t			-> (match-first :s :: t)

definition	= _ identifier:name plusequal expression: expr semicolon?	-> (prepend-rule :name         :expr)
		| _ identifier:name equalplus expression: expr semicolon?	-> (append-rule  :name         :expr)
		| _ identifier:name field*:fields equal
			expression:expr semicolon?				-> (define-rule  :name :fields :expr)

grammar		= definition*

start		= grammar
