/* primitive.c -- primitive functions
** 
** Copyright (c) 2009 Ian Piumarta
** All Rights Reserved
** 
** Permission is hereby granted, free of charge, to any person obtaining a
** copy of this software and associated documentation files (the 'Software'),
** to deal in the Software without restriction, including without limitation
** the rights to use, copy, modify, merge, publish, distribute, and/or sell
** copies of the Software, and to permit persons to whom the Software is
** furnished to do so, provided that the above copyright notice(s) and this
** permission notice appear in all copies of the Software.  Inclusion of the
** the above copyright notice(s) and this permission notice in supporting
** documentation would be appreciated but is not required.
** 
** THE SOFTWARE IS PROVIDED 'AS IS'.  USE ENTIRELY AT YOUR OWN RISK.
** 
** Last edited: 2010-08-03 11:14:31 by piumarta on emilia-2.local
*/

#if OLD_ABI
# define arglist cdr
#else
# define arglist
#endif


static oop Subr_environment(oop args, oop env)
{
  return env;
}


static oop Subr_error(oop args, oop env)
{
  printf("\n");
  for (args= arglist(args);  isPair(args);  args= getTail(args)) {
    print(getHead(args));
    if (isPair(getTail(args))) printf(" ");
  }
  printf("\n");
  abort();
  return 0;
}


static oop Subr_trace(oop args, oop env)
{
  args= car(arglist(args));
  optTrace= isLong(args) ? getLong(args) : 0;
  return nil;
}


static oop Subr_properties(oop args, oop env)
{
  args= arglist(args);
  return getProperties(car(args));
}

/* static oop lassq(oop key, oop list) */
/* { */
/*   for (;  isPair(list);  list= getTail(list)) */
/*     if (eq(key, car(getHead(list)))) */
/*       return list; */
/*   return nil; */
/* } */

static oop Subr_properties_at(oop args, oop env)
{
  args= arglist(args);
  oop obj= car(args);
  oop key= cadr(args);
  if (isNil(obj) || isLong(obj) || isNil(key)) return nil;
  oop props= getProperties(obj);
  oop ptr, prev= 0;
  for (ptr= props;  isPair(ptr);  ptr= getTail(prev= ptr)) {
    oop assoc= getHead(ptr);
    if (key == getHead(assoc)) {
      if (prev) {
	oop temp= getHead(prev);
	set(prev, Pair,head, assoc);
	set(ptr,  Pair,head, temp);
      }
      return getTail(assoc);
    }
  }
  return nil;
}

static oop Subr_set_properties_at(oop args, oop env)
{
  args= arglist(args);
  oop obj=   car(args);
  oop key=   cadr(args);
  oop value= caddr(args);
  oop props= getProperties(obj);
  if (isNil(obj) || isLong(obj) || isNil(key)) return nil;
  if (notNil(value)) {
    oop assoc= assq(key, props);
    if (isNil(assoc)) {
      assoc= mkPair(key, value);			GC_PROTECT(assoc);
      setProperties(obj, mkPair(assoc, props));		GC_UNPROTECT(assoc);
      return value;
    }
    return rplacd(assoc, value);
  }
  if (eq(key, caar(props))) {
    setProperties(obj, cdr(props));
    return value;
  }
  oop next;
  for (;  isPair(next= cdr(props));  props= next) {
    if (eq(key, caar(next))) {
      rplacd(props, cdr(next));
      return value;
    }
  }
  return value;
}


static oop Subr_read(oop args, oop env)
{
  args= arglist(args);
  oop arg= car(args);
  oop rnv= cadr(args);		if (isNil(env)) rnv= env;
  INPUT *stream= 0;
  switch (getTypeIndex(arg)) {
  case iLong:
    stream= (INPUT *)getLong(arg);
    break;
  case iPointer:
    stream= (INPUT *)get(arg, Pointer,address);
    break;
  default:
    return nil;
  }
  arg= readObject(stream, rnv);
  return (arg == readerEOF) ? nil : arg;
}


static oop applyEnvironment(oop function)
{
  switch (getTypeIndex(function)) {
  case iExpr:	return get(function, Expr,env);
  case iSyntax: return applyEnvironment(get(function, Syntax,form));
  }
  return nil;
}

static oop Subr_apply(oop args, oop env)
{
  args= arglist(args);
  oop function= car(args);
  oop arguments= cadr(args);
  oop environment= caddr(args);
  if (isNil(environment)) environment= applyEnvironment(function);
  return apply(function, arguments, environment);
}

static oop Subr_eval(oop args, oop env)
{
  args= arglist(args);
  oop expression= car(args);
  oop environment= cadr(args);
  return eval(expression, environment);
}

static oop Subr_exec(oop args, oop env)
{
  args= arglist(args);
  oop program= car(args);
  oop arguments= cadr(args);
  oop environment= caddr(args);
  if (isNil(environment)) environment= env;
  return exec(program, arguments, environment);
}


static oop Subr_dump(oop args, oop env)
{
  args= arglist(args);
  dump("", car(args));
  return nil;
}

static oop Subr_print(oop args, oop env)
{
  oop result= nil;
  for (args= arglist(args);  isPair(args);  args= getTail(args)) {
    print(result= car(args));
    if (isPair(getTail(args))) printf(" ");
  }
  return result;
}


static oop Subr_float(oop args, oop env)
{
  args= arglist(args);
  oop arg= car(args);
  if (isFloat(arg)) return arg;
  if (isLong(arg)) return mkFloat((double)getLong(arg));
  return error("non-numeric argument in float");  
}

static oop Subr_floatP(oop args, oop env)
{
  args= arglist(args);
  return mkBoolean(isFloat(car(args)));
}


static void finalise(void *ptr, void *data)
{
  GC_PROTECT(ptr);
  GC_PROTECT(data);
  oop args= cons(ptr, nil);
  GC_PROTECT(args);
  apply((oop)data, args, applyEnvironment(data));
  GC_UNPROTECT(ptr);
}

static oop Subr_finalise(oop args, oop env)
{
  args= arglist(args);
  oop ptr= car(args);
  oop fun= cadr(args);
  if (!isExpr(fun) && !isPointer(fun) && !isSubr(fun))
    return error("non-applicable function in finalise");
  GC_register_finaliser(ptr, finalise, fun);
  return ptr;
}


static oop Subr_pointer(oop args, oop env)
{
  args= arglist(args);
  return mkPointer(car(args), nil, nil);
}

static oop Subr_pointerP(oop args, oop env)
{
  args= arglist(args);
  return mkBoolean(isPointer(car(args)));
}

static oop Subr_pointer_address(oop args, oop env)
{
  args= arglist(args);
  oop ptr= car(args);				if (!isPointer(ptr)) return error("non-pointer argument in pointer-address");
  return mkLong((long)get(ptr, Pointer,address));
}

static oop Subr_set_pointer_address(oop args, oop env)
{
  args= arglist(args);
  oop ptr= car(args);				if (!isPointer(ptr))	return error("non-pointer argument in set-pointer-address");
  oop adr= cadr(args);				if (!isLong(adr))	return error("non-integer argument in set-pointer-address");
  set(ptr, Pointer,address, (void *)getLong(adr));
  return adr;
}


static oop Subr_syntax(oop args, oop env)
{
  args= arglist(args);
  return mkSyntax(car(args));
}

static oop Subr_syntaxP(oop args, oop env)
{
  args= arglist(args);
  return mkBoolean(isSyntax(car(args)));
}

static oop Subr_syntax_function(oop args, oop env)
{
  args= arglist(args);
  oop syntax= car(args);			if (!isSyntax(syntax)) return error("non-syntax argument in syntax-function");
  return get(syntax, Syntax,compiler);
}

static oop Subr_set_syntax_function(oop args, oop env)
{
  args= arglist(args);
  oop syntax= car(args);			if (!isSyntax(syntax)) return error("non-syntax argument in syntax-function");
  oop function= cadr(args);
  return set(syntax, Syntax,compiler, function);
}


static oop makeFields(oop fields, int index)
{
  if (!isPair(fields)) return nil;
  oop rest=   makeFields(getTail(fields), index + sizeof(oop));		GC_PROTECT(rest);
  oop name=   car(fields);
  oop get=    s_oopAt;
  oop offset= mkLong(index);
  oop size=   mkLong(4);
  if (isPair(name)) {
    get= cadr(name);
    if (isNil(get)) get= s_ptrAt;
    name= car(name);
  }
  oop field= cons(offset, size);					GC_PROTECT(field);
  field= cons(get, field);
  field= cons(name, field);
  rest=  cons(field, rest);						GC_UNPROTECT(rest);
  return rest;
}

static oop Subr_new_type(oop args, oop env)
{
  args= arglist(args);
  oop  name  = car(args);			if (!(isSymbol(name) || isString(name))) return error("non-symbol name in new-type");
  oop  fields= makeFields(cadr(args), 0);
  long size  = sizeof(oop) * list_length(fields);
  return mkType(name, fields, size, mark_type, print_type);
}

static oop Subr_new(oop args, oop env)
{
  args= arglist(args);
  oop     type=   car(args);		if (!(isType(type))) return error("non-type in new");
  long    size=   get(type, Type,size);
  int     index=  get(type, Type,index);
  oop     obj=    allocPointers(index, size);
  oop    *fields= getBody(obj);
  long    limit=  size /= sizeof(oop);
  for (size= 0;  size < limit;  ++size)  fields[size]= nil;
  return obj;
}

static oop Subr_typeof(oop args, oop env)
{
  args= arglist(args);
  return getType(car(args));
}

static oop Subr_type_index(oop args, oop env)
{
  args= arglist(args);
  oop type= car(args);
  if (!isType(type)) error("non-type argument in type-index");
  return mkLong(get(type, Type,index));
}

static oop Subr_type_name(oop args, oop env)
{
  args= arglist(args);
  oop type= car(args);
  if (!isType(type)) error("non-type argument in type-name");
  return get(type, Type,name);
}

static oop Subr_type_fields(oop args, oop env)
{
  args= arglist(args);
  oop type= car(args);
  if (!isType(type)) error("non-type argument in type-fields");
  return get(type, Type,fields);
}

static char *print_l2(char *buf, long size, long l)
{
  if (l > 9) buf= print_l2(buf, size, l / 10);
  *buf= '0' + (l % 10);
  return buf + 1;
}

static char *print_l(char *buf, long size, long l)
{
  if (l < 0) {
    *buf= '-';
    buf= print_l2(buf+1, size-1, -l);
  }
  else
    buf= print_l2(buf, size, l);
  *buf= '\0';
  return buf;
}

static char *print_f3(char *buf, long size, double d)
{
  double di, df;
  df= modf(d, &di);
  buf= print_l2(buf, size, di);
  *buf++= '.';
  buf= print_l2(buf, size, df * 1000.0 + 0.5);
  return buf;
}

static char *print_f2(char *buf, long size, double d)
{
  double dl= log10(d);
  if (dl > 6 || dl < -3) {
    double dm;
    dm= d * pow(10.0, -dl);
    buf= print_f3(buf, size, dm);
    *buf++= 'e';
    buf= print_l(buf, size, dl);
  }
  else {
    buf= print_f3(buf, size, d);
  }
  *buf= '\0';
  return buf;
}
  
static void print_f(char *buf, long size, double d)
{
  if (d < 0.0) {
    *buf= '-';
    print_f2(buf+1, size-1, -d);
  }
  else
    print_f2(buf, size, d);
}

static oop Subr_long(oop args, oop env)
{
  args= arglist(args);
  oop arg= car(args);
  switch (getTypeIndex(arg)) {
  case iUndefined:	return mkLong(0);
  case iTag:		/* ... */
  case iLong:		return arg;
  case iFloat:		return mkLong((long)getFloat(arg));
  case iPointer:	return mkLong((long)get(arg, Pointer,address));
  case iPair:		break;
  case iString:		return mkLong((long)atoi(get(arg, String,elements)));
  case iSymbol:		break;
  case iArray:		break;
  case iByteArray:	break;
  case iWordArray:	break;
  case iSyntax:		break;
  case iExpr:		break;
  case iSubr:		break;
  case iType:		break;
  case iContext:	break;
  }
  return nil;
}

static oop Subr_longP(oop args, oop env)
{
  args= arglist(args);
  return mkBoolean(isLong(car(args)));
}

static oop Subr_string(oop args, oop env)
{
  args= arglist(args);
  oop arg= car(args);
  switch (getTypeIndex(arg)) {
  case iUndefined:	return mkStringN(0);
  case iTag:		/* ... */
  case iLong: {
    long n= 0;
    for (arg= args;  isPair(arg) && isLong(getHead(arg));  arg= getTail(arg)) ++n;
    oop string= mkStringN(n + 1);
    long i;
    for (i= 0;  i < n;  ++i, args= getTail(args))
      get(string, String,elements)[i]= getLong(car(args));
    set(string, String,size, mkLong(n));
    return string;
  }
  case iFloat:		{ char buf[32];  print_f(buf, 32, getFloat(arg));  return mkString(buf); }
  case iPointer:	return mkString((char *)get(arg, Pointer,address));
  case iPair:		break;
  case iString:		return mkString((char *)get(arg, String,elements));
  case iSymbol:		return mkString((char *)get(get(arg, Symbol,string), String,elements));
  case iArray:		return nil;
  case iByteArray:	return mkString((char *)get(arg, ByteArray,elements));
  case iWordArray:	break;
  case iSyntax:		break;
  case iExpr:		break;
  case iSubr:		break;
  case iType:		return get(get(arg, Type,name), Symbol,string);
  case iContext:	break;
  }
  return nil;
}

static oop Subr_stringP(oop args, oop env)
{
  args= arglist(args);
  return mkBoolean(isString(car(args)));
}

static oop Subr_string_length(oop args, oop env)
{
  args= arglist(args);
  oop  string= car(args);			if (!isString(string))	return error("non-string argument in string-length");
  return get(string, String,size);
}

static oop Subr_string_at(oop args, oop env)
{
  args= arglist(args);
  oop  string= car(args);	if (!isString(string))	return error("non-string argument in string-at");
  oop  index= cadr(args);	if (!isLong(index))	return error("non-integer index in string-at");
  long i= getLong(index);	if (i < 0)		return error("negative index in string-at");
  long s= getLong(get(string, String,size));
  if (i < s) return mkLong(((unsigned char *)get(string, String,elements))[i]);
  return caddr(args);
}

static oop Subr_set_string_at(oop args, oop env)
{
  args= arglist(args);
  oop  string= car(args);	if (!isString(string))	return error("non-string argument in set-string-at");
  oop  index= cadr(args);	if (!isLong(index))	return error("non-integer index in set-string-at");
  oop  value= caddr(args);	if (!isLong(value))	return error("non-integer value in set-string-at");
  long i=     getLong(index);	if (i < 0)		return error("negative index in set-string-at");
  long size=  getLong(get(string, String,size));
  byte *elements= (byte *)get(string, String,elements);
  if (size <= i) {
    long j=   i + 1;
    long cap= GC_size(elements);
    if (cap <= j) {
      long newCap= cap * 2;
      while (newCap <= j) newCap *= 2;
      elements= GC_realloc(elements, sizeof(byte) * newCap);
      set(string, String,elements, (char *)elements);
    }
    while (size <= i) elements[size++]= '\0';
    elements[j]= '\0';
    set(string, String,size, mkLong(size));
  }
  elements[i]= getLong(value);
  return value;
}


static oop Subr_symbol(oop args, oop env)
{
  args= arglist(args);
  oop arg= car(args);
  switch (getTypeIndex(arg)) {
  case iPointer:	return mkSymbol((char *)get(arg, Pointer,address));
  case iString:		return mkSymbol((char *)get(arg, String,elements));
  case iSymbol:		return arg;
  case iByteArray:	return mkSymbol((char *)get(arg, ByteArray,elements));
  case iType:		return get(arg, Type,name);
  default:		break;
  }
  return nil;
}

static oop Subr_symbolP(oop args, oop env)
{
  args= arglist(args);
  return mkBoolean(isSymbol(car(args)));
}

static oop Subr_symbol_opcode(oop args, oop env)
{
  args= arglist(args);
  oop symbol= car(args);		if (!isSymbol(symbol))	{ println(symbol); return error("non-symbol argument in symbol-opcode"); }
  return mkLong(get(symbol, Symbol,opcode));
}


static oop Subr_array(oop args, oop env)
{
  args= arglist(args);
  oop argp;
  long size= 0;
  for (argp= args;  isPair(argp);  argp= cdr(argp)) ++size;
  oop array= mkArray(size);
  size= 0;
  for (argp= args;  isPair(argp);  argp= cdr(argp)) get(array, Array,elements)[size++]= car(argp);
  return array;
}

static oop Subr_array_length(oop args, oop env)
{
  args= arglist(args);
  oop  array= car(args);	if (!isArray(array))	return error("non-array argument in array-length");
  return get(array, Array,size);
}

static oop Subr_array_swap(oop args, oop env)
{
  args= arglist(args);
  oop  arr1= car(args);		if (!isArray(arr1))	return error("non-array argument 1 in array-swap");
  oop  arr2= cadr(args);	if (!isArray(arr2))	return error("non-array argument 2 in array-swap");
  oop  tmp= mkArray(0);		GC_PROTECT(tmp);

  set(tmp,  Array,elements, 0);
  set(tmp,  Array,size,     get(arr1, Array,size));
  set(tmp,  Array,elements, get(arr1, Array,elements));
  set(arr1, Array,elements, 0);
  set(arr1, Array,size,     get(arr2, Array,size));
  set(arr1, Array,elements, get(arr2, Array,elements));
  set(arr2, Array,elements, 0);
  set(arr2, Array,size,     get(tmp, Array,size));
  set(arr2, Array,elements, get(tmp, Array,elements));
  GC_UNPROTECT(tmp);
  return arr1;
}

static oop Subr_array_at(oop args, oop env)
{
  args= arglist(args);
  oop  array= car(args);	if (!isArray(array))	return error("non-array argument in array-at");
  oop  index= cadr(args);	if (!isLong(index))	return error("non-integer index in array-at");
  long i= getLong(index);	if (i < 0)		return error("negative index in array-at");
  long s= getLong(get(array, Array,size));
  if (i < s) return get(array, Array,elements)[i];
  return get(array, Array,undefined);
}

static oop Subr_array_at_typeof(oop args, oop env)
{
  args= arglist(args);
  oop  array= car(args);	if (!isArray(array))	return error("non-array argument in array-at");
  oop  index= cadr(args);
  long i= getTypeIndex(index);
  long s= getLong(get(array, Array,size));
  if (i < s) {
    oop elt= get(array, Array,elements)[i];
    if (notNil(elt)) return elt;
  }
  return caddr(args);
}

static oop Subr_set_array_at(oop args, oop env)
{
  args= arglist(args);
  oop  array= car(args);	if (!isArray(array))	return error("non-array argument in set-array-at");
  oop  index= cadr(args);	if (!isLong(index))	return error("non-integer index in set-array-at");
  oop  value= caddr(args);
  long i=     getLong(index);	if (i < 0)		return error("negative index in set-array-at");
  long size=  getLong(get(array, Array,size));
  oop *elements= get(array, Array,elements);
  if (i >= size) {
    long cap= GC_size(elements) / sizeof(oop);
    if (i >= cap) {
      long newCap= cap * 2;
      while (newCap <= i) newCap *= 2;
      elements= GC_realloc(elements, sizeof(oop) * newCap);
      set(array, Array,elements, elements);
    }
    oop undef= get(array, Array,undefined);
    while (size <= i) elements[size++]= undef;
    set(array, Array,size, mkLong(size));
  }
  return elements[i]= value;
}


static oop Subr_word_array(oop args, oop env)
{
  args= arglist(args);
  oop argp;
  long size= 0;
  for (argp= args;  isPair(argp);  argp= cdr(argp)) ++size;
  oop array= mkWordArray(size);
  size= 0;
  for (argp= args;  isPair(argp);  argp= cdr(argp)) get(array, WordArray,elements)[size++]= objectToWord(car(argp));
  return array;
}

static oop Subr_word_array_length(oop args, oop env)
{
  args= arglist(args);
  oop  array= car(args);	if (!isWordArray(array))	return error("non-word-array argument in word-array-length");
  return get(array, WordArray,size);
}

static oop Subr_word_array_at(oop args, oop env)
{
  args= arglist(args);
  oop  array= car(args);	if (!isWordArray(array))	return error("non-array argument in word-array-at");
  oop  index= cadr(args);	if (!isLong(index))		return error("non-integer index in word-array-at");
  long i= getLong(index);	if (i < 0)			return error("negative index in word-array-at");
  long s= getLong(get(array, WordArray,size));
  if (i < s) return mkPointer((void *)get(array, WordArray,elements)[i], nil, nil);
  return caddr(args);
}

static oop Subr_set_word_array_at(oop args, oop env)
{
  args= arglist(args);
  oop  array= car(args);	if (!isWordArray(array))	return error("non-array argument in set-word-array-at");
  oop  index= cadr(args);	if (!isLong(index))		return error("non-integer index in set-word-array-at");
  oop  value= caddr(args);
  long i=     getLong(index);	if (i < 0)			return error("negative index in set-word-array-at");
  long size=  getLong(get(array, WordArray,size));
  word *elements= get(array, WordArray,elements);
  if (i >= size) {
    long cap= GC_size(elements) / sizeof(word);
    if (i >= cap) {
      long newCap= cap * 2;
      while (newCap <= i) newCap *= 2;
      elements= GC_realloc(elements, sizeof(word) * newCap);
      set(array, WordArray,elements, elements);
    }
    while (size <= i) elements[size++]= nil;
    set(array, WordArray,size, mkLong(size));
  }
  elements[i]= objectToWord(value);
  return value;
}


static oop Subr_byte_array(oop args, oop env)
{
  args= arglist(args);
  long size= 0;
  oop  arg=  car(args);
  if (notNil(arg)) {
    if (!isLong(arg)) return error("non-integer argument in byte-array");
    size= getLong(arg);
  }
  return mkByteArray(size);
}

static oop Subr_byte_array_length(oop args, oop env)
{
  args= arglist(args);
  oop  array= car(args);	if (!isByteArray(array))	return error("non-byte-array argument in byte-array-length");
  return get(array, ByteArray,size);
}

static oop Subr_byte_array_at(oop args, oop env)
{
  args= arglist(args);
  oop  array= car(args);	if (!isByteArray(array))	return error("non-array argument in byte-array-at");
  oop  index= cadr(args);	if (!isLong(index))		return error("non-integer index in byte-array-at");
  long i= getLong(index);	if (i < 0)			return error("negative index in byte-array-at");
  long s= getLong(get(array, ByteArray,size));
  if (i < s) return mkLong(get(array, ByteArray,elements)[i]);
  return caddr(args);
}

static oop Subr_set_byte_array_at(oop args, oop env)
{
  args= arglist(args);
  oop  array= car(args);	if (!isByteArray(array))	return error("non-array argument in set-byte-array-at");
  oop  index= cadr(args);	if (!isLong(index))		return error("non-integer index in set-byte-array-at");
  oop  value= caddr(args);	if (!isLong(index))		return error("non-integer value in set-byte-array-at");
  long i=     getLong(index);	if (i < 0)			return error("negative index in set-byte-array-at");
  long size=  getLong(get(array, ByteArray,size));
  byte *elements= get(array, ByteArray,elements);
  if (i >= size) {
    long cap= GC_size(elements);
    if (i >= cap) {
      long newCap= cap * 2;
      while (newCap <= i) newCap *= 2;
      elements= GC_realloc(elements, sizeof(byte) * newCap);
      set(array, ByteArray,elements, elements);
    }
    while (size <= i) elements[size++]= 0;
    set(array, ByteArray,size, mkLong(size));
  }
  elements[i]= getLong(value);
  return value;
}


static oop Subr_dlopen(oop args, oop env)
{
  args= arglist(args);
  oop symbol= car(args);	if (!(isSymbol(symbol) || isString(symbol))) return error("non-symbol argument in dlopen");
  char *name= tostring(symbol);
  void *handle= dlopen(name, RTLD_NOW | RTLD_GLOBAL);
  return handle ? mkPointer(handle, symbol, nil) : nil;
}

static oop Subr_extern(oop args, oop env)
{
  args= arglist(args);
  oop 	symbol= car(args);	if (!(isSymbol(symbol) || isString(symbol))) return error("non-symbol argument in extern");
  oop 	type=  cadr(args);
  oop 	result= nil;
  GC_PROTECT(result);
  char *name= tostring(symbol);
  void *addr= dlsym(RTLD_DEFAULT, name);
  char *err;
  if (isNil(type)) {
    warning("type of '%s' defaults to (* void)", tostring(symbol));
    type= t_pointer_void;
  }
  err= dlerror();
  if (!err)
    result= mkPointer(addr, symbol, type);
  else
    fprintf(stderr, "cannot resolve `%s': %s\n", name, err);
  GC_UNPROTECT(result);
  return result;
}

static oop Subr_externP(oop args, oop env)
{
  args= arglist(args);
  return mkBoolean(isPointer(car(args)));
}


static oop Subr_primitiveP(oop args, oop env)
{
  args= arglist(args);
  return mkBoolean(isSubr(car(args)));
}


static oop Subr_function(oop args, oop env)
{
  args= arglist(args);
  return mkExpr(car(args), cadr(args), caddr(args), cadddr(args), caddddr(args));
}

static oop Subr_functionP(oop args, oop env)
{
  args= arglist(args);
  return mkBoolean(isExpr(car(args)));
}

static oop Subr_function_parameters(oop args, oop env)
{
  args= arglist(args);
  oop func= car(args);
  if (!isExpr(func)) return error("non-function in function-parameters");
  return get(func, Expr,params);
}

static oop Subr_function_body(oop args, oop env)
{
  args= arglist(args);
  oop func= car(args);
  if (!isExpr(func)) return error("non-function in function-body");
  return get(func, Expr,body);
}

static oop Subr_function_method(oop args, oop env)
{
  args= arglist(args);
  oop func= car(args);
  if (!isExpr(func)) return error("non-function in function-method");
  return get(func, Expr,method);
}

static oop Subr_set_function_method(oop args, oop env)
{
  args= arglist(args);
  oop func= car(args);
  oop method= cadr(args);
  if (!isExpr(func)) return error("non-function in set-function-method");
  return set(func, Expr,method, method);
}

static oop Subr_function_environment(oop args, oop env)
{
  args= arglist(args);
  oop func= car(args);
  if (!isExpr(func)) return error("non-function in function-environment");
  return get(func, Expr,env);
}


static oop Subr_context(oop args, oop env)
{
  args= arglist(args);
  oop  sa= cadr(args);	if (!isArray(sa)) return error("non-array stack in context");
  word sz= getLong(get(sa, Array,size));
  oop  cx= mkContextN(sz);
  oop  ip= caddddr(args);
  oop *sp= get(cx, Context,stackBase);
  set(cx, Context,pc,         car(args));
  set(cx, Context,env,        caddr(args));
  set(cx, Context,caller,     cadddr(args));
  set(cx, Context,kernel,     cadddddr(args));
  set(cx, Context,ip,         isLong(ip) ? getLong(ip) : 0);
  set(cx, Context,sp,         sp);
  memcpy(sp, get(sa, Array,elements), sizeof(oop) * sz);
  return cx;
}

static oop Subr_contextP(oop args, oop env)
{
  args= arglist(args);
  return mkBoolean(isContext(car(args)));
}

static oop Subr_context_pc(oop args, oop env)
{
  args= arglist(args);
  oop context= car(args);
  if (!isContext(context)) return error("non-context in context-pc");
  return get(context, Context,pc);
}

static oop Subr_context_stack(oop args, oop env)
{
  args= arglist(args);
  oop context= car(args);
  if (!isContext(context)) return error("non-context in context-stack");
  oop *sptr= get(context, Context,sp);
  oop *slim= get(context, Context,stackLimit);
  word size= slim - sptr;
  oop  sarr= mkArray(size);
  oop *elts= get(sarr, Array,elements);
  word i;
  for (i= 0;  i < size;  ++i) elts[i]= sptr[i];
  return sarr;
}

static oop Subr_context_env(oop args, oop env)
{
  args= arglist(args);
  oop context= car(args);
  if (!isContext(context)) return error("non-context in context-env");
  return get(context, Context,env);
}

static oop Subr_set_context_env(oop args, oop env)
{
  args= arglist(args);
  oop context= car(args);
  oop value= cadr(args);
  if (!isContext(context)) return error("non-context in context-env");
  return set(context, Context,env, value);
}

static oop Subr_context_caller(oop args, oop env)
{
  args= arglist(args);
  oop context= car(args);
  if (!isContext(context)) return error("non-context in context-caller");
  return get(context, Context,caller);
}

static oop Subr_context_ip(oop args, oop env)
{
  args= arglist(args);
  oop context= car(args);
  if (!isContext(context)) return error("non-context in context-ip");
  return mkLong(get(context, Context,ip));
}

static oop Subr_context_kernel(oop args, oop env)
{
  args= arglist(args);
  oop context= car(args);
  if (!isContext(context)) return error("non-context in context-kernel");
  return get(context, Context,kernel);
}

static oop Subr_set_context_kernel(oop args, oop env)
{
  args= arglist(args);
  oop context= car(args);
  oop kernel= cadr(args);
  if (!isContext(context))					return error("non-context in set-context-kernel");
  if (!isArray(kernel))						return error("non-Array kernel in set-context-kernel");
  if (getLong(get(kernel, Array,size)) < CONTEXT_KERNEL_END)	return error("wrong size kernel in set-context-kernel");
  return set(context, Context,kernel, kernel);
}


static oop concat(oop a, oop b)
{
  if (isNil(b)) return a;
  if (isNil(a)) return b;
  switch (getTypeIndex(a)) {
  case iUndefined: {
    if (!isPair(b)) break;
    return b;
  }
  case iSymbol: {
    if (!isSymbol(b)) break;
    oop   sa=  get(a, Symbol,string);
    oop   sb=  get(b, Symbol,string);
    long  la=  getLong(get(sa, String,size));
    long  lb=  getLong(get(sb, String,size));
    char *buf= alloca(la + lb + 1);
    strcpy(buf+ 0, get(sa, String,elements));
    strcpy(buf+la, get(sb, String,elements));
    return mkSymbol(buf);
  }
  case iString: {
    if (!isString(b)) break;
    long  la=  getLong(get(a, String,size));
    long  lb=  getLong(get(b, String,size));
    char *buf= alloca(la + lb + 1);
    strcpy(buf+ 0, get(a, String,elements));
    strcpy(buf+la, get(b, String,elements));
    return mkString(buf);
  }
  case iPair: {
    oop list= isPair(b) ? list_concat(b, nil) : b;	GC_PROTECT(list);
    list= list_concat(a, list);				GC_UNPROTECT(list);
    return list;
  }
  default:
    break;
  }
  return error("cannot concat %s with %s", tostring(get(getType(a), Type,name)), tostring(get(getType(b), Type,name)));
}

static oop Subr_concat(oop args, oop env)
{
  args= arglist(args);
  oop result= nil;
  GC_PROTECT(result);
  for (result= car(args), args= cdr(args);  isPair(args);  args= getTail(args)) {
    result= concat(result, getHead(args));
  }
  GC_UNPROTECT(result);
  return result;
}


static oop Subr_longAt(oop args, oop env)
{
  args= arglist(args);
  oop  object= car(args);
  oop  index=  cadr(args);		if (!isLong(index))	return error("non-integer index in long@");
  long offset= getLong(index);
  void *base=  0;
  switch (getTypeIndex(object)) {
  case iLong:		base= (void *)getLong(object);		break;
  case iPointer:	base= get(object, Pointer,address);	break;
  default:		base= getBody(object);			break;
  }
  return mkLong(*(long *)(base + offset));
}

static oop Subr_set_longAt(oop args, oop env)
{
  args= arglist(args);
  oop  object= car(args);
  oop  index=  cadr(args);		if (!isLong(index))	return error("non-integer index in set-long@");
  oop  value=  caddr(args);		if (!isLong(value))	return error("non-integer value in set-long@");
  long offset= getLong(index);
  void *base=  0;
  switch (getTypeIndex(object)) {
  case iLong:		base= (void *)getLong(object);		break;
  case iPointer:	base= get(object, Pointer,address);	break;
  default:		base= getBody(object);			break;
  }
  *(long *)(base + offset)= getLong(value);
  return value;
}

static oop Subr_setIf_longAt(oop args, oop env)
{
  args= arglist(args);
  oop  object= car(args);
  oop  index=  cadr(args);		if (!isLong(index))	return error("non-integer index in set-if-long@");
  oop  oldval= caddr(args);		if (!isLong(oldval))	return error("non-integer old value in set-if-long@");
  oop  newval= cadddr(args);		if (!isLong(newval))	return error("non-integer new value in set-if-long@");
  long offset= getLong(index);
  void *base=  0;
  switch (getTypeIndex(object)) {
  case iLong:		base= (void *)getLong(object);		break;
  case iPointer:	base= get(object, Pointer,address);	break;
  default:		base= getBody(object);			break;
  }
  long *eaddr= (long *)(base + offset);
  if (getLong(oldval) != *eaddr) {
    oop timeout= caddddr(args);
    if (isNil(timeout))
      return error("deadlock detected in set-if-long@");
    return nil;
  }
  *eaddr= getLong(newval);
  return t;
}

static oop Subr_oopAt(oop args, oop env)
{
  args= arglist(args);
  oop  object= car(args);
  oop  index=  cadr(args);		if (!isLong(index))	return error("non-integer index in oop@");
  long offset= getLong(index);
  return *(oop *)(void *)(((char *)getBody(object)) + offset);
}

static oop Subr_set_oopAt(oop args, oop env)
{
  args= arglist(args);
  oop  object= car(args);
  oop  index=  cadr(args);	if (!isLong(index))	return error("non-integer index in set-oop@");
  oop  value=  caddr(args);
  long offset= getLong(index);
  return *(oop *)(void *)(((char *)getBody(object)) + offset)= value;
}

static oop Subr_setIf_oopAt(oop args, oop env)
{
  args= arglist(args);
  oop  object= car(args);
  oop  index=  cadr(args);		if (!isLong(index))	return error("non-integer index in set-if-oop@");
  oop  oldval= caddr(args);
  oop  newval= cadddr(args);
  long offset= getLong(index);
  oop *eaddr= (oop *)((char *)getBody(object) + offset);
  if (oldval != *eaddr) {
    oop timeout= caddddr(args);
    if (isNil(timeout))
      return error("deadlock detected in set-if-oop@");
    return nil;
  }
  *eaddr= newval;
  return t;
}


static oop Subr_cons(oop args, oop env)
{
  args= arglist(args);
  return cons(car(args), cadr(args));
}

static oop Subr_pairP(oop args, oop env)
{
  args= arglist(args);
  return mkBoolean(isPair(car(args)));
}

static oop Subr_list_length(oop args, oop env)
{
  args= arglist(args);
  return mkLong(list_length(car(args)));
}

static oop Subr_car(oop args, oop env)
{
  args= arglist(args);
  return caar(args);
}

static oop Subr_set_car(oop args, oop env)
{
  args= arglist(args);
  oop cell= car(args);
  oop value= cadr(args);
  return isPair(cell) ? rplaca(cell, value) : cell;
}

static oop Subr_cdr(oop args, oop env)
{
  args= arglist(args);
  return cdar(args);
}

static oop Subr_set_cdr(oop args, oop env)
{
  args= arglist(args);
  oop cell= car(args);
  oop value= cadr(args);
  return isPair(cell) ? rplacd(cell, value) : cell;
}


static oop Subr_assoc(oop args, oop env)
{
  args= arglist(args);
  oop key= car(args);
  oop alist= cadr(args);
  return assoc(key, alist);
}

static oop Subr_assq(oop args, oop env)
{
  args= arglist(args);
  oop key= car(args);
  oop alist= cadr(args);
  return assq(key, alist);
}

static oop Subr_nssq(oop args, oop env)
{
  args= arglist(args);
  oop key= car(args);
  oop alist= cadr(args);
  return nssq(key, alist);
}


static oop Subr_not(oop args, oop env)
{
  args= arglist(args);
  return mkBoolean(isNil(car(args)));
}


#define arithmetic(ID, OP)									\
static oop Subr_##ID(oop args, oop env)								\
{												\
  args= arglist(args);										\
  oop lhs= car(args);										\
  oop rhs= cadr(args);										\
  if (isLong(lhs) && isLong(rhs)) return mkLong(getLong(lhs) OP getLong(rhs));			\
  switch (getTypeIndex(lhs)) {									\
  case iLong:											\
    switch (getTypeIndex(rhs)) {								\
    case iLong:		return mkLong(getLong(lhs) OP getLong(rhs));				\
    case iFloat:	return mkFloat(getLong(lhs) OP getFloat(rhs));				\
    default:		break;									\
    }												\
    break;											\
  case iFloat:											\
    switch (getTypeIndex(rhs)) {								\
    case iLong:		return mkFloat(getFloat(lhs) OP getLong(rhs));				\
    case iFloat:	return mkFloat(getFloat(lhs) OP getFloat(rhs));				\
    default:		break;									\
    }												\
    break;											\
  default:											\
    break;											\
  }												\
  fprintf(stderr, "\n\n");  fprintln(args, stderr);						\
  return error("non-numeric argument (%d %d) in "#OP, getTypeIndex(lhs), getTypeIndex(rhs));	\
}

arithmetic(add,	+)
arithmetic(sub,	-)
arithmetic(mul,	*)
arithmetic(div,	/)

#undef arithmetic

static oop Subr_mod(oop args, oop env)
{
  args= arglist(args);
  oop lhs= car(args);
  oop rhs= cadr(args);
  switch (getTypeIndex(lhs)) {
  case iLong:
    switch (getTypeIndex(rhs)) {
    case iLong:		return mkLong((long)getLong(lhs) % (long)getLong(rhs));
    case iFloat:	return mkLong((long)getLong(lhs) % (long)getFloat(rhs));
    default:		break;
    }
  case iFloat:
    switch (getTypeIndex(rhs)) {
    case iLong:		return mkLong((long)getFloat(lhs) % (long)getLong(rhs));
    case iFloat:	return mkLong((long)getFloat(lhs) % (long)getFloat(rhs));
    default:		break;
    }
    break;
  default:
    break;
  }
  return error("non-numeric argument (%d %d) in %%", getTypeIndex(lhs), getTypeIndex(rhs));
}

#define bitwise(ID, OP)									\
static oop Subr_##ID(oop args, oop env)							\
{											\
  args= arglist(args);									\
  oop lhs= car(args);									\
  oop rhs= cadr(args);									\
  if (!isLong(lhs) || !isLong(rhs)) return error("non-integer argument in "#OP);	\
  return mkLong(getLong(lhs) OP getLong(rhs));						\
}

bitwise(and,  &)
bitwise(or,   |)
bitwise(xor,  ^)
bitwise(asl, <<)
bitwise(asr, >>)

#undef bitwise

#define relation(ID, OP)									\
static oop Subr_##ID(oop args, oop env)								\
{												\
  args= arglist(args);										\
  oop lhs= car(args);										\
  oop rhs= cadr(args);										\
  if (isLong(lhs) && isLong(rhs)) return mkBoolean(getLong(lhs) OP getLong(rhs));		\
  switch (getTypeIndex(lhs)) {									\
  case iLong:											\
    switch (getTypeIndex(rhs)) {								\
    case iLong:		return mkBoolean(getLong(lhs) OP getLong(rhs));				\
    case iFloat:	return mkBoolean(getLong(lhs) OP getFloat(rhs));			\
    default:		break;									\
    }												\
    break;											\
  case iFloat:											\
    switch (getTypeIndex(rhs)) {								\
    case iLong:		return mkBoolean(getFloat(lhs) OP getLong(rhs));			\
    case iFloat:	return mkBoolean(getFloat(lhs) OP getFloat(rhs));			\
    default:		break;									\
    }												\
    break;											\
  default:											\
    break;											\
  }												\
  return error("non-numeric argument (%d %d) in "#OP, getTypeIndex(lhs), getTypeIndex(rhs));	\
}

relation(lt,	<)
relation(le,	<=)
relation(ge,	>=)
relation(gt,	>)

#undef relation

#define relation(ID, OP, NAME)							\
static oop Subr_##ID(oop args, oop env)						\
{										\
  args= arglist(args);								\
  oop a= car(args);								\
  oop b= cadr(args);								\
  return mkBoolean(OP(a, b));							\
}

relation(equal,	  equal, =)
relation(nequal, !equal, !=)
relation(eq,	  eq,    ==)
relation(neq,	 !eq,    !==)

#undef relation

/**************** platform-specific primitives ****************/

# if defined(OFW)

static inline unsigned char inb(unsigned short port)
{
  unsigned char _v;
  asm("inb %w1,%0" : "=a"(_v) : "Nd"(port));
  return _v;
}

static oop Subr_inb(oop args)
{
  args= arglist(args);
  oop arg= car(args);		if (!isLong(arg)) return error("non-numeric argument in inb");
  return mkLong(inb(getLong(arg)));
}

static inline void outb(unsigned char value, unsigned short port)
{
  asm("outb %b0,%w1" : : "a"(value), "Nd"(port));
}

static oop Subr_outb(oop args)
{
  args= arglist(args);
  oop arg1= car(args);		if (!isLong(arg1)) return error("non-numeric argument in outb");
  oop arg2= cadr(args);		if (!isLong(arg2)) return error("non-numeric argument in outb");
  outb(getLong(arg1), getLong(arg2));
  return arg1;
}

#else

static oop Subr_inb(oop args)
{
  return error("inb not implemented on this platform");
}

static oop Subr_outb(oop args)
{
  return error("outb not implemented on this platform");
}

#endif

#undef arglist
