/* types.c -- build descriptors for the builtin types
** 
** Copyright (c) 2009 Ian Piumarta
** All Rights Reserved
** 
** Permission is hereby granted, free of charge, to any person obtaining a
** copy of this software and associated documentation files (the 'Software'),
** to deal in the Software without restriction, including without limitation
** the rights to use, copy, modify, merge, publish, distribute, and/or sell
** copies of the Software, and to permit persons to whom the Software is
** furnished to do so, provided that the above copyright notice(s) and this
** permission notice appear in all copies of the Software.  Inclusion of the
** the above copyright notice(s) and this permission notice in supporting
** documentation would be appreciated but is not required.
** 
** THE SOFTWARE IS PROVIDED 'AS IS'.  USE ENTIRELY AT YOUR OWN RISK.
** 
** Last edited: 2010-08-03 11:14:28 by piumarta on emilia-2.local
*/

static void makeTypes(void)
{
  tType= mkType(nil, nil, sizeof(Type), mark_Type, print_Type);
  tType->_type= iType;
  GC_add_root(&tType);
  assert(isNil(types));
  assert(-1 == get(tType, Type,index));
# define t(T) {								\
    t##T= mkType(nil, nil, sizeof(T), mark_##T, print_##T);		\
    GC_add_root(&t##T);							\
    assert(-1 == get(t##T, Type,index));				\
  }
  typesDo(t);
  types= mkArray(0);
# undef t
# define t(T) {								\
    set(t##T, Type,index, getLong(get(types, Array,size)));		\
    array_append(types, t##T);						\
  }
  t(Type);  typesDo(t);
# undef t
  assert(iPair == get(tPair, Type,index));
# define t(T) set(t##T, Type,name, mkSymbol(#T));
  t(Type);  typesDo(t);
# undef t
}

#if 1
static void initType(oop type, ...)
{
  set(type, Type,fields, nil);
  va_list ap;
  va_start(ap, type);
  oop fields= nil;		GC_PROTECT(fields);
  oop desc= nil;		GC_PROTECT(desc);
  long offset= 0;
  oop getter;
  char *name;
  while ((getter= va_arg(ap, oop)) && (name= va_arg(ap, char *))) {
    desc= cons(nil, nil);
    rplaca(desc, mkLong(offset));
    rplacd(desc, mkLong(sizeof(oop)));
    desc= cons(nil, desc);
    rplaca(desc, getter);
    desc= cons(nil, desc);
    rplaca(desc, mkSymbol(name));
    fields= cons(desc, fields);
    offset += sizeof(oop);
  }
  while (isPair(fields)) {
    oop field= fields;
    fields= getTail(fields);
    rplacd(field, get(type, Type,fields));
    set(type, Type,fields, field);
  }
  GC_UNPROTECT(fields);
  va_end(ap);
}
#endif

static void initTypes(void)
{
  oop s_longAt	= mkSymbol("long@");
  oop s_floatAt	= mkSymbol("float@");
  initType(tType,      s_longAt,  "index",    s_oopAt,  "name",     s_oopAt, "fields",    s_longAt, "size",     s_ptrAt, "mark",   s_ptrAt, "print",                    0);
  initType(tLong,      s_longAt,  "value",                                                                                                                              0);
  initType(tFloat,     s_floatAt, "value",                                                                                                                              0);
  initType(tPointer,   s_ptrAt,   "address",  s_oopAt,  "name",     s_oopAt, "proto",                                                                                   0);
  initType(tPair,      s_oopAt,   "head",     s_oopAt,  "tail",                                                                                                         0);
  initType(tString,    s_longAt,  "capacity", s_longAt, "size",     s_ptrAt, "elements",                                                                                0);
  initType(tSymbol,    s_oopAt,   "string",                                                                                                                             0);
  initType(tArray,     s_longAt,  "capacity", s_longAt, "size",     s_oopAt, "undefined", s_ptrAt,  "elements",                                                         0);
  initType(tByteArray, s_longAt,  "capacity", s_longAt, "size",     s_oopAt, "elements",                                                                                0);
  initType(tWordArray, s_longAt,  "capacity", s_longAt, "size",     s_oopAt, "elements",                                                                                0);
  initType(tSyntax,    s_oopAt,   "form",     s_oopAt,  "compiler",                                                                                                     0);
  initType(tExpr,      s_oopAt,   "expr",     s_oopAt,  "params",   s_oopAt, "body",      s_oopAt,  "method",   s_oopAt, "env",    s_oopAt, "name",                     0);
  initType(tSubr,      s_ptrAt,   "exec",     s_oopAt,  "name",                                                                                                         0);
  initType(tContext,   s_oopAt,   "pc",       s_oopAt,  "stack",    s_oopAt, "env",       s_oopAt,  "caller",   s_oopAt, "callee", s_oopAt, "function", s_longAt, "ip", 0);
}
